import copy
import fnmatch
import json
import os
from json import JSONDecodeError
from typing import Iterable, Tuple, Dict

from conan.api.model import RecipeReference, PkgReference
from conan.api.output import ConanOutput
from conan.errors import ConanException
from conan.internal.errors import NotFoundException
from conan.internal.model.version_range import VersionRange
from conan.internal.graph.graph import RECIPE_EDITABLE, RECIPE_CONSUMER, RECIPE_PLATFORM, \
    RECIPE_VIRTUAL, BINARY_SKIP, BINARY_MISSING, BINARY_INVALID
from conan.internal.util.files import load


class MultiPackagesList:
    """ A collection of PackagesList by remote name."""
    def __init__(self):
        self.lists = {}

    def __getitem__(self, name):
        try:
            return self.lists[name]
        except KeyError:
            raise ConanException(f"'{name}' doesn't exist in package list")

    def add(self, name, pkg_list):
        self.lists[name] = pkg_list

    def add_error(self, remote_name, error):
        self.lists[remote_name] = {"error": error}

    def serialize(self):
        """ Serialize object to a dictionary."""
        return {k: v.serialize() if isinstance(v, PackagesList) else v
                for k, v in self.lists.items()}

    def merge(self, other):
        for k, v in other.lists.items():
            self.lists.setdefault(k, PackagesList()).merge(v)

    def keep_outer(self, other):
        for namespace, other_pkg_list in other.lists.items():
            self.lists.get(namespace, PackagesList()).keep_outer(other_pkg_list)

    @staticmethod
    def load(file):
        """ Create an instance of the class from a serialized JSON file path pointed by ``file``."""
        try:
            content = json.loads(load(file))
        except JSONDecodeError as e:
            raise ConanException(f"Package list file invalid JSON: {file}\n{e}")
        except Exception as e:
            raise ConanException(f"Package list file missing or broken: {file}\n{e}")
        # Check if input json is not a graph file
        if "graph" in content:
            base_path = os.path.basename(file)
            raise ConanException(
                'Expected a package list file but found a graph file. You can create a "package list" JSON file by running:\n\n'
                f"\tconan list --graph {base_path} --format=json > pkglist.json\n\n"
                "More Info at 'https://docs.conan.io/2/examples/commands/pkglists.html"
            )
        result = {}
        for remote, pkglist in content.items():
            if "error" in pkglist:
                result[remote] = pkglist
            else:
                result[remote] = PackagesList.deserialize(pkglist)
        pkglist = MultiPackagesList()
        pkglist.lists = result
        return pkglist

    @staticmethod
    def load_graph(graphfile, graph_recipes=None, graph_binaries=None, context=None):
        """ Create an instance of the class from a graph file path, which is
        the json format returned by a few commands
        like ``conan graph info`` or ``conan create/install.``

        :parameter str graphfile: Path to the graph file
        :parameter list[str] graph_recipes: List for kinds of recipes to return.
            For example ``"cache"`` will return only recipes in the local cache,
            ``"download"`` will return only recipes that have been downloaded,
            and passing ``"*"`` will return all recipes.
        :parameter list[str] graph_binaries: List for kinds of binaries to return.
            For example ``"cache"`` will return only binaries in the local cache,
            ``"download"`` will return only binaries that have been downloaded,
            ``"build"`` will return only binaries that are built,
            ``"missing"`` will return only binaries that are missing,
            ``"invalid"`` will return only binaries that are invalid,
            and passing ``"*"`` will return all binaries.
        :parameter str context: Context to filter the graph,
            can be ``"host"``, ``"build"``, ``"host-only"`` or ``"build-only"``
        """
        if not os.path.isfile(graphfile):
            raise ConanException(f"Graph file not found: {graphfile}")
        try:
            base_context = context.split("-")[0] if context else None
            graph = json.loads(load(graphfile))
            # Check if input json is a graph file
            if "graph" not in graph:
                raise ConanException(
                    'Expected a graph file but found an unexpected JSON file format. You can create a "graph" JSON file by running:\n\n'
                    f"\tconan [ graph-info | create | export-pkg | install | test ] --format=json > graph.json\n\n"
                    "More Info at 'https://docs.conan.io/2/reference/commands/formatters/graph_info_json_formatter.html"
                )

            mpkglist = MultiPackagesList._define_graph(graph, graph_recipes, graph_binaries,
                                                       context=base_context)
            if context == "build-only":
                host = MultiPackagesList._define_graph(graph, graph_recipes, graph_binaries,
                                                       context="host")
                mpkglist.keep_outer(host)
            elif context == "host-only":
                build = MultiPackagesList._define_graph(graph, graph_recipes, graph_binaries,
                                                        context="build")
                mpkglist.keep_outer(build)
            return mpkglist
        except JSONDecodeError as e:
            raise ConanException(f"Graph file invalid JSON: {graphfile}\n{e}")
        except KeyError as e:
            raise ConanException(f'Graph file {graphfile} is missing the required "{e}" key in its contents.\n'
                                 "Note that the graph file should not be filtered "
                                 "if you expect to use it with the list command.")
        except ConanException as e:
            raise e
        except Exception as e:
            raise ConanException(f"Graph file broken: {graphfile}\n{e}")

    @staticmethod
    def _define_graph(graph, graph_recipes=None, graph_binaries=None, context=None):
        pkglist = MultiPackagesList()
        cache_list = PackagesList()
        if graph_recipes is None and graph_binaries is None:
            recipes = ["*"]
            binaries = ["*"]
        else:
            recipes = [r.lower() for r in graph_recipes or []]
            binaries = [b.lower() for b in graph_binaries or []]

        pkglist.lists["Local Cache"] = cache_list
        for node in graph["graph"]["nodes"].values():
            if context and node['context'] != context:
                continue

            # We need to add the python_requires too
            python_requires = node.get("python_requires")
            if python_requires is not None:
                for pyref, pyreq in python_requires.items():
                    pyrecipe = pyreq["recipe"]
                    if pyrecipe == RECIPE_EDITABLE:
                        continue
                    pyref = RecipeReference.loads(pyref)
                    if any(r == "*" or r == pyrecipe for r in recipes):
                        cache_list.add_ref(pyref)
                    pyremote = pyreq["remote"]
                    if pyremote:
                        remote_list = pkglist.lists.setdefault(pyremote, PackagesList())
                        remote_list.add_ref(pyref)

            recipe = node["recipe"]
            if recipe in (RECIPE_EDITABLE, RECIPE_CONSUMER, RECIPE_VIRTUAL, RECIPE_PLATFORM):
                continue

            ref = node["ref"]
            ref = RecipeReference.loads(ref)
            ref.timestamp = node["rrev_timestamp"]
            recipe = recipe.lower()
            if any(r == "*" or r == recipe for r in recipes):
                cache_list.add_ref(ref)

            remote = node["remote"]
            if remote:
                remote_list = pkglist.lists.setdefault(remote, PackagesList())
                remote_list.add_ref(ref)
            pref = PkgReference(ref, node["package_id"], node["prev"], node["prev_timestamp"])
            binary_remote = node["binary_remote"]
            if binary_remote:
                remote_list = pkglist.lists.setdefault(binary_remote, PackagesList())
                remote_list.add_ref(ref)  # Binary listed forces recipe listed
                remote_list.add_pref(pref)

            binary = node["binary"]
            if binary in (BINARY_SKIP, BINARY_INVALID, BINARY_MISSING):
                continue

            binary = binary.lower()
            if any(b == "*" or b == binary for b in binaries):
                cache_list.add_ref(ref)  # Binary listed forces recipe listed
                cache_list.add_pref(pref, node["info"])
        return pkglist


class PackagesList:
    """ A collection of recipes, revisions and packages."""
    def __init__(self):
        self._data = {}

    def __bool__(self):
        """ Whether the package list contains any recipe"""
        return bool(self._data)

    def merge(self, other):
        assert isinstance(other, PackagesList)

        def recursive_dict_update(d, u):  # TODO: repeated from conandata.py
            for k, v in u.items():
                if isinstance(v, dict):
                    d[k] = recursive_dict_update(d.get(k, {}), v)
                else:
                    d[k] = v
            return d
        recursive_dict_update(self._data, other._data)

    def keep_outer(self, other):
        assert isinstance(other, PackagesList)
        if not self._data:
            return

        for ref, info in other._data.items():
            if self._data.get(ref, {}) == info:
                self._data.pop(ref)

    def split(self):
        """
        Returns a list of PackageList, split one per reference.
        This can be useful to parallelize things like upload, parallelizing per-reference
        """
        result = []
        for r, content in self._data.items():
            subpkglist = PackagesList()
            subpkglist._data[r] = content
            result.append(subpkglist)
        return result

    def only_recipes(self) -> None:
        """ Filter out all the packages and package revisions, keep only the recipes and
            recipe revisions in self._data.
        """
        for ref, ref_dict in self._data.items():
            for rrev_dict in ref_dict.get("revisions", {}).values():
                rrev_dict.pop("packages", None)

    def add_refs(self, refs):
        ConanOutput().warning("PackagesLists.add_refs() non-public, non-documented method will be "
                              "removed, use .add_ref() instead", warn_tag="deprecated")
        # RREVS alreday come in ASCENDING order, so upload does older revisions first
        for ref in refs:
            self.add_ref(ref)

    def add_ref(self, ref: RecipeReference) -> None:
        """
        Adds a new RecipeReference to a package list
        """
        ref_dict = self._data.setdefault(str(ref), {})
        if ref.revision:
            revs_dict = ref_dict.setdefault("revisions", {})
            rev_dict = revs_dict.setdefault(ref.revision, {})
            if ref.timestamp:
                rev_dict["timestamp"] = ref.timestamp

    def add_prefs(self, rrev, prefs):
        ConanOutput().warning("PackageLists.add_prefs() non-public, non-documented method will be "
                              "removed, use .add_pref() instead", warn_tag="deprecated")
        # Prevs already come in ASCENDING order, so upload does older revisions first
        for p in prefs:
            self.add_pref(p)

    def add_pref(self, pref: PkgReference, pkg_info: dict = None) -> None:
        """
        Add a PkgReference to an already existing RecipeReference inside a package list
        """
        # Prevs already come in ASCENDING order, so upload does older revisions first
        rev_dict = self.recipe_dict(pref.ref)
        packages_dict = rev_dict.setdefault("packages", {})
        package_dict = packages_dict.setdefault(pref.package_id, {})
        if pref.revision:
            prevs_dict = package_dict.setdefault("revisions", {})
            prev_dict = prevs_dict.setdefault(pref.revision, {})
            if pref.timestamp:
                prev_dict["timestamp"] = pref.timestamp
        if pkg_info is not None:
            package_dict["info"] = pkg_info

    def add_configurations(self, confs):
        ConanOutput().warning("PackageLists.add_configurations() non-public, non-documented method "
                              "will be removed, use .add_pref() instead",
                              warn_tag="deprecated")
        for pref, conf in confs.items():
            rev_dict = self.recipe_dict(pref.ref)
            try:
                rev_dict["packages"][pref.package_id]["info"] = conf
            except KeyError:  # If package_id does not exist, do nothing, only add to existing prefs
                pass

    def refs(self):
        ConanOutput().warning("PackageLists.refs() non-public, non-documented method will be "
                              "removed, use .items() instead", warn_tag="deprecated")
        result = {}
        for ref, ref_dict in self._data.items():
            for rrev, rrev_dict in ref_dict.get("revisions", {}).items():
                t = rrev_dict.get("timestamp")
                recipe = RecipeReference.loads(f"{ref}#{rrev}")  # TODO: optimize this
                if t is not None:
                    recipe.timestamp = t
                result[recipe] = rrev_dict
        return result

    def items(self) -> Iterable[Tuple[RecipeReference, Dict[PkgReference, Dict]]]:
        """ Iterate the contents of the package list.

        The first dictionary is the information directly belonging to the recipe-revision.
        The second dictionary contains PkgReference as keys, and a dictionary with the values
        belonging to that specific package reference (settings, options, etc.).
        """
        for ref, ref_dict in self._data.items():
            for rrev, rrev_dict in ref_dict.get("revisions", {}).items():
                recipe = RecipeReference.loads(f"{ref}#{rrev}")  # TODO: optimize this
                t = rrev_dict.get("timestamp")
                if t is not None:
                    recipe.timestamp = t
                packages = {}
                for package_id, pkg_info in rrev_dict.get("packages", {}).items():
                    prevs = pkg_info.get("revisions", {})
                    for prev, prev_info in prevs.items():
                        t = prev_info.get("timestamp")
                        pref = PkgReference(recipe, package_id, prev, t)
                        packages[pref] = prev_info
                yield recipe, packages

    def recipe_dict(self, ref: RecipeReference):
        """ Gives read/write access to the dictionary containing a specific RecipeReference
        information.
        """
        return self._data[str(ref)]["revisions"][ref.revision]

    def package_dict(self, pref: PkgReference):
        """ Gives read/write access to the dictionary containing a specific PkgReference
        information
        """
        ref_dict = self.recipe_dict(pref.ref)
        return ref_dict["packages"][pref.package_id]["revisions"][pref.revision]

    @staticmethod
    def prefs(ref, recipe_bundle):
        ConanOutput().warning("PackageLists.prefs() non-public, non-documented method will be "
                              "removed, use .items() instead", warn_tag="deprecated")
        result = {}
        for package_id, pkg_bundle in recipe_bundle.get("packages", {}).items():
            prevs = pkg_bundle.get("revisions", {})
            for prev, prev_bundle in prevs.items():
                t = prev_bundle.get("timestamp")
                pref = PkgReference(ref, package_id, prev, t)
                result[pref] = prev_bundle
        return result

    def serialize(self):
        """ Serialize the instance to a dictionary."""
        return copy.deepcopy(self._data)

    @staticmethod
    def deserialize(data):
        """ Loads the data from a serialized dictionary."""
        result = PackagesList()
        result._data = copy.deepcopy(data)
        return result


class ListPattern:
    """ Object holding a pattern that matches recipes, revisions and packages."""

    def __init__(self, expression, rrev="latest", package_id=None, prev="latest", only_recipe=False):
        """
        :param expression: The pattern to match, e.g. ``"name/*:*"``
        :param rrev: The recipe revision to match, defaults to ``"latest"``,
                     can also be ``"!latest"`` or ``"~latest"`` to match all but the latest revision,
                     a pattern like ``"1234*"`` to match a specific revision,
                     or a specific revision like ``"1234"``.
        :param package_id: The package ID to match, defaults to ``None``, which matches all package IDs.
        :param prev: The package revision to match, defaults to ``"latest"``,
                     can also be ``"!latest"`` or ``"~latest"`` to match all but the latest revision,
                     a pattern like ``"1234*"`` to match a specific revision,
                     or a specific revision like ``"1234"``.
        :param only_recipe: If ``True``, only the recipe part of the expression is parsed,
                            ignoring ``package_id`` and ``prev``. This is useful for commands that
                            only operate on recipes, like ``conan search``.
        """
        def split(s, c, default=None):
            if not s:
                return None, default
            tokens = s.split(c, 1)
            if len(tokens) == 2:
                return tokens[0], tokens[1] or default
            return tokens[0], default

        recipe, package = split(expression, ":")
        self.raw = expression
        self.ref, rrev = split(recipe, "#", rrev)
        ref, user_channel = split(self.ref, "@")
        self.name, self.version = split(ref, "/")
        self.user, self.channel = split(user_channel, "/")
        self.rrev, _ = split(rrev, "%")
        self.package_id, prev = split(package, "#", prev)
        self.prev, _ = split(prev, "%")
        if only_recipe:
            if self.package_id:
                raise ConanException("Do not specify 'package_id' with 'only-recipe'")
        else:
            self.package_id = self.package_id or package_id

    @staticmethod
    def _only_latest(rev):
        return rev in ["!latest", "~latest"]

    @property
    def search_ref(self):
        vrange = self._version_range
        if vrange:
            return str(RecipeReference(self.name, "*", self.user, self.channel))
        if "*" in self.ref or not self.version or (self.package_id is None and self.rrev is None):
            return self.ref

    @property
    def _version_range(self):
        if self.version and self.version.startswith("[") and self.version.endswith("]"):
            return VersionRange(self.version[1:-1])

    def filter_versions(self, refs, resolve_prereleases=None):
        vrange = self._version_range
        if vrange:
            refs = [r for r in refs if vrange.contains(r.version, resolve_prereleases)]
        return refs

    @property
    def is_latest_rrev(self):
        return self.rrev == "latest"

    @property
    def is_latest_prev(self):
        return self.prev == "latest"

    def check_refs(self, refs):
        if not refs and self.ref and "*" not in self.ref:
            raise NotFoundException(f"Recipe '{self.ref}' not found")

    def filter_rrevs(self, rrevs):
        if self._only_latest(self.rrev):
            return rrevs[1:]
        rrevs = [r for r in rrevs if fnmatch.fnmatch(r.revision, self.rrev)]
        if not rrevs:
            refs_str = f'{self.ref}#{self.rrev}'
            if "*" not in refs_str:
                raise ConanException(f"Recipe revision '{refs_str}' not found")
        return rrevs

    def filter_prefs(self, prefs):
        prefs = [p for p in prefs if fnmatch.fnmatch(p.package_id, self.package_id)]
        if not prefs:
            refs_str = f'{self.ref}#{self.rrev}:{self.package_id}'
            if "*" not in refs_str:
                raise ConanException(f"Package ID '{self.raw}' not found")
        return prefs

    def filter_prevs(self, prevs):
        if self._only_latest(self.prev):
            return prevs[1:]
        prevs = [p for p in prevs if fnmatch.fnmatch(p.revision, self.prev)]
        if not prevs:
            refs_str = f'{self.ref}#{self.rrev}:{self.package_id}#{self.prev}'
            if "*" not in refs_str:
                raise ConanException(f"Package revision '{self.raw}' not found")
        return prevs



import textwrap

from conan.test.utils.tools import TestClient


def test_cross_build_options():
    # https://github.com/conan-io/conan/issues/8443
    c = TestClient()
    dep = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "dep"
            version = "0.1"
            options = {"fPIC": [True, False]}
            default_options = {"fPIC": True}
            settings = "os"
            def config_options(self):
                if self.settings.os == "Windows":
                    del self.options.fPIC
            """)
    consumer = textwrap.dedent("""
        from conan import ConanFile
        class Consumer(ConanFile):
            requires = "dep/0.1"
            tool_requires = "dep/0.1"
        """)
    c.save({"dep/conanfile.py": dep,
            "consumer/conanfile.py": consumer})
    c.run("create dep -s os=Android -s os.api_level=22")
    c.run("create dep -s os=Windows")
    c.run("install consumer -s:b os=Windows -s:h os=Android -s:h os.api_level=22")
    # The above command used to crash, because options not there, so it works now without problems
    assert "Finalizing install" in c.out
    c.assert_listed_binary({"dep/0.1": ("4333c2a38a8baf674c4e77fe475c146a4685f77a", "Cache")})
    c.assert_listed_binary({"dep/0.1": ("ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715", "Cache")},
                           build=True)



import pytest
from unittest import mock

from conan.tools.build import check_max_cppstd, check_min_cppstd, valid_max_cppstd, valid_min_cppstd, \
    valid_min_cstd, valid_max_cstd
from conan.test.utils.mocks import MockSettings, ConanFileMock
from conan.errors import ConanException, ConanInvalidConfiguration


def _create_conanfile(compiler, version, os, cppstd, libcxx=None):
    settings = MockSettings({"arch": "x86_64",
                             "build_type": "Debug",
                             "os": os,
                             "compiler": compiler,
                             "compiler.version": version,
                             "compiler.cppstd": cppstd})
    if libcxx:
        settings.values["compiler.libcxx"] = libcxx
    conanfile = ConanFileMock(settings)
    return conanfile


class TestUserInput:

    def test_check_cppstd_type(self):
        """ cppstd must be a number
        """
        conanfile = ConanFileMock(MockSettings({}))
        with pytest.raises(ConanException) as raises:
            check_min_cppstd(conanfile, "gnu17", False)
        assert "cppstd parameter must be a number" == str(raises.value)


class TestCheckMinCppStd:

    @pytest.mark.parametrize("cppstd", ["98", "11", "14", "17"])
    def test_check_min_cppstd_from_settings(self, cppstd):
        """ check_min_cppstd must accept cppstd less/equal than cppstd in settings
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "17", "libstdc++")
        check_min_cppstd(conanfile, cppstd, False)

    @pytest.mark.parametrize("cppstd", ["98", "11", "14"])
    def test_check_min_cppstd_from_outdated_settings(self, cppstd):
        """ check_min_cppstd must raise when cppstd is greater when supported on settings
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", cppstd, "libstdc++")
        with pytest.raises(ConanInvalidConfiguration) as raises:
            check_min_cppstd(conanfile, "17", False)
        assert "Current cppstd ({}) is lower than the required C++ standard (17).".format(cppstd) == str(raises.value)

    @pytest.mark.parametrize("cppstd", ["98", "11", "14", "17"])
    def test_check_min_cppstd_from_settings_with_extension(self, cppstd):
        """ current cppstd in settings must has GNU extension when extensions is enabled
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "gnu17", "libstdc++")
        check_min_cppstd(conanfile, cppstd, True)

        conanfile.settings.values["compiler.cppstd"] = "17"
        with pytest.raises(ConanException) as raises:
            check_min_cppstd(conanfile, cppstd, True)
        assert "The cppstd GNU extension is required" == str(raises.value)

    def test_check_min_cppstd_unsupported_standard(self):
        """ check_min_cppstd must raise when the compiler does not support a standard
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "gnu14", "libstdc++")
        with pytest.raises(ConanInvalidConfiguration) as raises:
            check_min_cppstd(conanfile, "42", False)
        assert "Current cppstd (gnu14) is lower than the required C++ standard (42)." == str(raises.value)

    def test_check_min_cppstd_gnu_compiler_extension(self):
        """ Current compiler must support GNU extension on Linux when extensions is required
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "17", "libstdc++")
        with mock.patch("platform.system", mock.MagicMock(return_value="Linux")):
            with pytest.raises(ConanException) as raises:
                check_min_cppstd(conanfile, "17", True)
            assert "The cppstd GNU extension is required" == str(raises.value)


class TestValidMinCppstd:

    @pytest.mark.parametrize("cppstd", ["98", "11", "14", "17"])
    def test_valid_min_cppstd_from_settings(self, cppstd):
        """ valid_min_cppstd must accept cppstd less/equal than cppstd in settings
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "17", "libstdc++")
        assert valid_min_cppstd(conanfile, cppstd, False)

    @pytest.mark.parametrize("cppstd", ["98", "11", "14"])
    def test_valid_min_cppstd_from_outdated_settings(self, cppstd):
        """ valid_min_cppstd returns False when cppstd is greater when supported on settings
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", cppstd, "libstdc++")
        assert not valid_min_cppstd(conanfile, "17", False)

    @pytest.mark.parametrize("cppstd", ["98", "11", "14", "17"])
    def test_valid_min_cppstd_from_settings_with_extension(self, cppstd):
        """ valid_min_cppstd must returns True when current cppstd in settings has GNU extension and
            extensions is enabled
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "gnu17", "libstdc++")
        assert valid_min_cppstd(conanfile, cppstd, True)

        conanfile.settings.values["compiler.cppstd"] = "17"
        assert not valid_min_cppstd(conanfile, cppstd, True)

    def test_valid_min_cppstd_unsupported_standard(self):
        """ valid_min_cppstd must returns False when the compiler does not support a standard
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "17", "libstdc++")
        assert not valid_min_cppstd(conanfile, "42", False)

    def test_valid_min_cppstd_gnu_compiler_extension(self):
        """ valid_min_cppstd must returns False when current compiler does not support GNU extension
            on Linux and extensions is required
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "gnu1z", "libstdc++")
        with mock.patch("platform.system", mock.MagicMock(return_value="Linux")):
            assert not valid_min_cppstd(conanfile, "20", True)

    @pytest.mark.parametrize("cppstd", ["98", "11", "14", "17"])
    def test_min_cppstd_mingw_windows(self, cppstd):
        """ GNU extensions HAS effect on Windows when running a cross-building for Linux
        """
        with mock.patch("platform.system", mock.MagicMock(return_value="Windows")):
            conanfile = _create_conanfile("gcc", "9", "Linux", "gnu17", "libstdc++")
            assert valid_min_cppstd(conanfile, cppstd, True)

            conanfile.settings.values["compiler.cppstd"] = "17"
            assert not valid_min_cppstd(conanfile, cppstd, True)


class TestCheckMaxCppStd:

    @pytest.mark.parametrize("cppstd", ["98", "11", "14", "17"])
    def test_check_max_cppstd_from_settings(self, cppstd):
        """ check_max_cppstd must accept cppstd higher/equal than cppstd in settings
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "98", "libstdc++")
        check_max_cppstd(conanfile, cppstd, False)

    @pytest.mark.parametrize("cppstd", ["11", "14", "17"])
    def test_check_max_cppstd_from_outdated_settings(self, cppstd):
        """ check_max_cppstd must raise when cppstd is higher when supported on settings
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", cppstd, "libstdc++")
        with pytest.raises(ConanInvalidConfiguration) as raises:
            check_max_cppstd(conanfile, "98", False)
        assert "Current cppstd ({}) is higher than the required C++ standard (98).".format(cppstd) == str(raises.value)

    @pytest.mark.parametrize("cppstd", ["98", "11", "14", "17"])
    def test_check_max_cppstd_from_settings_with_extension(self, cppstd):
        """ current cppstd in settings must have GNU extension when extensions is enabled
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "gnu98", "libstdc++")
        check_max_cppstd(conanfile, cppstd, True)

        conanfile.settings.values["compiler.cppstd"] = "98"
        with pytest.raises(ConanException) as raises:
            check_max_cppstd(conanfile, cppstd, True)
        assert "The cppstd GNU extension is required" == str(raises.value)

    def test_check_max_cppstd_unsupported_standard(self):
        """ check_max_cppstd must raise when the compiler does not support a standard
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "gnu17", "libstdc++")
        with pytest.raises(ConanInvalidConfiguration) as raises:
            check_max_cppstd(conanfile, "16", False)
        assert "Current cppstd (gnu17) is higher than the required C++ standard (16)." == str(raises.value)

    def test_check_max_cppstd_gnu_compiler_extension(self):
        """ Current compiler must support GNU extension on Linux when extensions is required
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "17", "libstdc++")
        with mock.patch("platform.system", mock.MagicMock(return_value="Linux")):
            with pytest.raises(ConanException) as raises:
                check_max_cppstd(conanfile, "17", True)
            assert "The cppstd GNU extension is required" == str(raises.value)


class TestValidMaxCppstd:

    @pytest.mark.parametrize("cppstd", ["98", "11", "14", "17"])
    def test_valid_max_cppstd_from_settings(self, cppstd):
        """ valid_max_cppstd must accept cppstd higher/equal than cppstd in settings
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "98", "libstdc++")
        assert valid_max_cppstd(conanfile, cppstd, False)

    @pytest.mark.parametrize("cppstd", ["11", "14", "17"])
    def test_valid_max_cppstd_from_outdated_settings(self, cppstd):
        """ valid_max_cppstd return False when cppstd is greater when supported on settings
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", cppstd, "libstdc++")
        assert not valid_max_cppstd(conanfile, "98", False)

    @pytest.mark.parametrize("cppstd", ["98", "11", "14", "17"])
    def test_valid_max_cppstd_from_settings_with_extension(self, cppstd):
        """ valid_max_cppstd must return True when current cppstd in settings has GNU extension and
            extensions is enabled
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "gnu98", "libstdc++")
        assert valid_max_cppstd(conanfile, cppstd, True)

        conanfile.settings.values["compiler.cppstd"] = "98"
        assert not valid_max_cppstd(conanfile, cppstd, True)

    def test_valid_max_cppstd_unsupported_standard(self):
        """ valid_max_cppstd must return False when the compiler does not support a standard
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "17", "libstdc++")
        assert not valid_max_cppstd(conanfile, "16", False)

    def test_valid_max_cppstd_gnu_compiler_extension(self):
        """ valid_max_cppstd must return False when current compiler does not support GNU extension
            on Linux and extensions is required
        """
        conanfile = _create_conanfile("gcc", "9", "Linux", "gnu17", "libstdc++")
        with mock.patch("platform.system", mock.MagicMock(return_value="Linux")):
            assert not valid_max_cppstd(conanfile, "14", True)

    @pytest.mark.parametrize("cppstd", ["98", "11", "14", "17"])
    def test_max_cppstd_mingw_windows(self, cppstd):
        """ GNU extensions HAS effect on Windows when running a cross-building for Linux
        """
        with mock.patch("platform.system", mock.MagicMock(return_value="Windows")):
            conanfile = _create_conanfile("gcc", "9", "Linux", "gnu98", "libstdc++")
            assert valid_max_cppstd(conanfile, cppstd, True)

            conanfile.settings.values["compiler.cppstd"] = "98"
            assert not valid_max_cppstd(conanfile, cppstd, True)


class TestCstd:

    @pytest.mark.parametrize("compiler, cstd, mincstd, valid",
                             [["gcc",  "99", "99", True],
                              ["gcc",  "gnu99", "99", True],
                              ["gcc", "99", "11", False]
                              ])
    def test_valid_min(self, compiler, cstd, mincstd, valid):
        settings = MockSettings({"arch": "x86_64",
                                 "build_type": "Debug",
                                 "os": "Linux",
                                 "compiler": compiler,
                                 "compiler.cstd": cstd})
        conanfile = ConanFileMock(settings)
        if valid:
            assert valid_min_cstd(conanfile, mincstd, False)
        else:
            assert not valid_min_cstd(conanfile, mincstd, False)

    @pytest.mark.parametrize("compiler, cstd, maxcstd, valid",
                             [["gcc", "99", "11", True],
                              ["gcc", "gnu99", "11", True],
                              ["gcc", "17", "11", False]
                              ])
    def test_valid_max(self, compiler, cstd, maxcstd, valid):
        settings = MockSettings({"arch": "x86_64",
                                 "build_type": "Debug",
                                 "os": "Linux",
                                 "compiler": compiler,
                                 "compiler.cstd": cstd})
        conanfile = ConanFileMock(settings)
        if valid:
            assert valid_max_cstd(conanfile, maxcstd, False)
        else:
            assert not valid_max_cstd(conanfile, maxcstd, False)

    def test_error(self):
        conanfile = ConanFileMock()
        with pytest.raises(ConanException) as e:
            valid_min_cstd(conanfile, "potato")
        assert "cstd parameter must be a number" in str(e.value)



import json
import os

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient
from conan.test.utils.env import environment_update


class TestOutputLevel:
    def test_invalid_output_level(self):
        t = TestClient(light=True)
        t.save({"conanfile.py": GenConanfile("foo", "1.0")})
        t.run("create . -vfooling", assert_error=True)
        assert "argument -v: invalid choice: 'fooling'" in t.out

        with environment_update({"CONAN_LOG_LEVEL": "fail"}):
            t.run("create .", assert_error=True)
            assert "Invalid argument '-vfail' defined in CONAN_LOG_LEVEL " \
                   "environment variable" in t.out

    def test_output_level(self):
        lines = ("self.output.trace('This is a trace')",
                 "self.output.debug('This is a debug')",
                 "self.output.verbose('This is a verbose')",
                 "self.output.info('This is a info')",
                 "self.output.highlight('This is a highlight')",
                 "self.output.success('This is a success')",
                 "self.output.warning('This is a warning')",
                 "self.output.error('This is a error')",
                 )

        t = TestClient(light=True)
        t.save({"conanfile.py": GenConanfile("foo", "1.0").with_package(*lines)})

        # By default, it prints > info
        t.run("create .")
        assert "This is a trace" not in t.out
        assert "This is a debug" not in t.out
        assert "This is a verbose" not in t.out
        assert "This is a info" in t.out
        assert "This is a highlight" in t.out
        assert "This is a success" in t.out
        assert "This is a warning" in t.out
        assert "This is a error" in t.out

        # Check if -v argument is equal to VERBOSE level
        t.run("create . -v")
        assert "This is a trace" not in t.out
        assert "This is a debug" not in t.out
        assert "This is a verbose" in t.out
        assert "This is a info" in t.out
        assert "This is a highlight" in t.out
        assert "This is a success" in t.out
        assert "This is a warning" in t.out
        assert "This is a error" in t.out

        # Print also verbose traces
        t.run("create . -vverbose")
        assert "This is a trace" not in t.out
        assert "This is a debug" not in t.out
        assert "This is a verbose" in t.out
        assert "This is a info" in t.out
        assert "This is a highlight" in t.out
        assert "This is a success" in t.out
        assert "This is a warning" in t.out
        assert "This is a error" in t.out

        # Print also debug traces
        t.run("create . -vv")
        assert "This is a trace" not in t.out
        assert "This is a debug" in t.out
        assert "This is a verbose" in t.out
        assert "This is a info" in t.out
        assert "This is a highlight" in t.out
        assert "This is a success" in t.out
        assert "This is a warning" in t.out
        assert "This is a error" in t.out
        t.run("create . -vdebug")
        assert "This is a trace" not in t.out
        assert "This is a debug" in t.out
        assert "This is a verbose" in t.out
        assert "This is a info" in t.out
        assert "This is a highlight" in t.out
        assert "This is a success" in t.out
        assert "This is a warning" in t.out
        assert "This is a error" in t.out

        # Print also "trace" traces
        t.run("create . -vvv")
        assert "This is a trace" in t.out
        assert "This is a debug" in t.out
        assert "This is a verbose" in t.out
        assert "This is a info" in t.out
        assert "This is a highlight" in t.out
        assert "This is a success" in t.out
        assert "This is a warning" in t.out
        assert "This is a error" in t.out
        t.run("create . -vtrace")
        assert "This is a trace" in t.out
        assert "This is a debug" in t.out
        assert "This is a verbose" in t.out
        assert "This is a info" in t.out
        assert "This is a highlight" in t.out
        assert "This is a success" in t.out
        assert "This is a warning" in t.out
        assert "This is a error" in t.out

        # With notice
        t.run("create . -vstatus")
        assert "This is a trace" not in t.out
        assert "This is a debug" not in t.out
        assert "This is a verbose" not in t.out
        assert "This is a info" in t.out
        assert "This is a highlight" in t.out
        assert "This is a success" in t.out
        assert "This is a warning" in t.out
        assert "This is a error" in t.out

        # With notice
        t.run("create . -vnotice")
        assert "This is a trace" not in t.out
        assert "This is a debug" not in t.out
        assert "This is a verbose" not in t.out
        assert "This is a info" not in t.out
        assert "This is a highlight" in t.out
        assert "This is a success" in t.out
        assert "This is a warning" in t.out
        assert "This is a error" in t.out

        # With warnings
        t.run("create . -vwarning")
        assert "This is a trace" not in t.out
        assert "This is a debug" not in t.out
        assert "This is a verbose" not in t.out
        assert "This is a info" not in t.out
        assert "This is a highlight" not in t.out
        assert "This is a success" not in t.out
        assert "This is a warning" in t.out
        assert "This is a error" in t.out

        # With errors
        t.run("create . -verror")
        assert "This is a trace" not in t.out
        assert "This is a debug" not in t.out
        assert "This is a verbose" not in t.out
        assert "This is a info" not in t.out
        assert "This is a highlight" not in t.out
        assert "This is a success" not in t.out
        assert "This is a warning" not in t.out
        assert "This is a error" in t.out


def test_output_level_envvar():
    lines = ("self.output.trace('This is a trace')",
             "self.output.debug('This is a debug')",
             "self.output.verbose('This is a verbose')",
             "self.output.info('This is a info')",
             "self.output.highlight('This is a highlight')",
             "self.output.success('This is a success')",
             "self.output.warning('This is a warning')",
             "self.output.error('This is a error')",
             )

    t = TestClient(light=True)
    t.save({"conanfile.py": GenConanfile().with_package(*lines)})

    # Check if -v argument is equal to VERBOSE level
    with environment_update({"CONAN_LOG_LEVEL": "verbose"}):
        t.run("create . --name foo --version 1.0")
        assert "This is a trace" not in t.out
        assert "This is a debug" not in t.out
        assert "This is a verbose" in t.out
        assert "This is a info" in t.out
        assert "This is a highlight" in t.out
        assert "This is a success" in t.out
        assert "This is a warning" in t.out
        assert "This is a error" in t.out

        # Check if -v argument is equal to VERBOSE level
    with environment_update({"CONAN_LOG_LEVEL": "error"}):
        t.run("create . --name foo --version 1.0")
        assert "This is a trace" not in t.out
        assert "This is a debug" not in t.out
        assert "This is a verbose" not in t.out
        assert "This is a info" not in t.out
        assert "This is a highlight" not in t.out
        assert "This is a success" not in t.out
        assert "This is a warning" not in t.out
        assert "This is a error" in t.out


class TestWarningHandling:
    warning_lines = ("self.output.warning('Tagged warning', warn_tag='tag')",
                     "self.output.warning('Untagged warning')")
    error_lines = ("self.output.error('Tagged error', error_type='exception')",
                   "self.output.error('Untagged error')")

    def test_warning_as_error_deprecated_syntax(self):
        t = TestClient(light=True)
        t.save({"conanfile.py": GenConanfile("foo", "1.0").with_package(*self.warning_lines)})

        t.save_home({"global.conf": "core:warnings_as_errors=[]"})
        t.run("create . -vwarning")
        assert "WARN: Untagged warning" in t.out
        assert "WARN: tag: Tagged warning" in t.out

        t.save_home({"global.conf": "core:warnings_as_errors=['*']"})
        t.run("create . -vwarning", assert_error=True)
        assert "ConanException: tag: Tagged warning" in t.out
        # We bailed early, didn't get a chance to print this one
        assert "Untagged warning" not in t.out

        t.save_home({"global.conf": """core:warnings_as_errors=['*']\ncore:skip_warnings=["tag"]"""})
        t.run("create . -verror", assert_error=True)
        assert "ConanException: Untagged warning" in t.out
        assert "Tagged warning" not in t.out

        t.save_home({"global.conf": "core:warnings_as_errors=[]"})
        t.run("create . -verror")
        assert "ERROR: Untagged warning" not in t.out
        assert "ERROR: tag: Tagged warning" not in t.out

    def test_skip_warnings(self):
        t = TestClient(light=True)
        t.save({"conanfile.py": GenConanfile("foo", "1.0").with_package(*self.warning_lines)})

        t.save_home({"global.conf": "core:skip_warnings=[]"})
        t.run("create . -vwarning")
        assert "WARN: Untagged warning" in t.out
        assert "WARN: tag: Tagged warning" in t.out

        t.save_home({"global.conf": "core:skip_warnings=['tag']"})
        t.run("create . -vwarning")
        assert "WARN: Untagged warning" in t.out
        assert "WARN: tag: Tagged warning" not in t.out

        t.save_home({"global.conf": "core:skip_warnings=['unknown']"})
        t.run("create . -vwarning")
        assert "WARN: Untagged warning" not in t.out
        assert "WARN: tag: Tagged warning" in t.out

        t.save_home({"global.conf": "core:skip_warnings=['unknown', 'tag']"})
        t.run("create . -vwarning")
        assert "WARN: Untagged warning" not in t.out
        assert "WARN: tag: Tagged warning" not in t.out

    def test_exception_errors(self):
        t = TestClient(light=True)
        t.save({"conanfile.py": GenConanfile("foo", "1.0").with_package(*self.error_lines)})

        t.save_home({"global.conf": "core:warnings_as_errors=[]"})
        t.run("create .")
        assert "ERROR: Tagged error" in t.out
        assert "ERROR: Untagged error" in t.out

        t.save_home({"global.conf": "core:warnings_as_errors=['*']"})
        t.run("create .", assert_error=True)
        assert "ERROR: Tagged error" in t.out
        assert "ConanException: Untagged error" in t.out

        t.run("create . -vquiet", assert_error=True)
        assert "ERROR: Tagged error" not in t.out
        assert "ConanException: Untagged error" not in t.out


def test_formatter_redirection_to_file():
    c = TestClient(light=True)
    c.save({"conanfile.py":  GenConanfile("pkg", "0.1")})

    c.run("config home --out-file=cmd_out.txt")
    assert "Formatted output saved to 'cmd_out.txt'" in c.out
    cmd_out = c.load("cmd_out.txt")
    assert f"{c.cache_folder}" in cmd_out
    assert not f"{c.cache_folder}" in c.stdout

    c.run("graph info . --format=json --out-file=graph.json")
    assert "Formatted output saved to 'graph.json'" in c.out
    graph = json.loads(c.load("graph.json"))
    assert len(graph["graph"]["nodes"]) == 1
    assert not "nodes" in c.stdout

    c.run("graph info . --format=html --out-file=graph.html")
    assert "Formatted output saved to 'graph.html'" in c.out
    html = c.load("graph.html")
    assert "<head>" in html
    assert not "<head>" in c.stdout

    c.run("install . --format=json --out-file=graph.json")
    assert "Formatted output saved to 'graph.json'" in c.out
    graph = json.loads(c.load("graph.json"))
    assert len(graph["graph"]["nodes"]) == 1
    assert not "nodes" in c.stdout


def test_redirect_to_file_create_dir():
    c = TestClient(light=True)
    c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
    c.run("config home --out-file=subdir/cmd_out.txt")
    cmd_out = c.load("subdir/cmd_out.txt")
    assert f"{c.cache_folder}" in cmd_out

    base = temp_folder()
    c = TestClient(light=True, current_folder=os.path.join(base, "sub1"))
    c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
    c.run("config home --out-file=../subdir/cmd_out.txt")
    cmd_out = c.load("../subdir/cmd_out.txt")
    assert f"{c.cache_folder}" in cmd_out



import textwrap
import pytest
import os

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def assert_vars_file(client, configuration):
    contents = client.load(f"conan_pkg.name-more+_vars_{configuration}_x86_64.premake5.lua")
    assert f'include "conanutils.premake5.lua"' in contents
    assert f't_conandeps = {{}}' in contents
    assert f't_conandeps["{configuration}_x86_64"] = {{}}' in contents
    assert f't_conandeps["{configuration}_x86_64"]["pkg.name-more+"] = {{}}' in contents
    assert f't_conandeps["{configuration}_x86_64"]["pkg.name-more+"]["includedirs"]' in contents
    assert f't_conandeps["{configuration}_x86_64"]["pkg.name-more+"]["libdirs"]' in contents
    assert f't_conandeps["{configuration}_x86_64"]["pkg.name-more+"]["bindirs"]' in contents
    assert f't_conandeps["{configuration}_x86_64"]["pkg.name-more+"]["libs"]' in contents
    assert f't_conandeps["{configuration}_x86_64"]["pkg.name-more+"]["system_libs"]' in contents
    assert f't_conandeps["{configuration}_x86_64"]["pkg.name-more+"]["defines"]' in contents
    assert f't_conandeps["{configuration}_x86_64"]["pkg.name-more+"]["cxxflags"]' in contents
    assert f't_conandeps["{configuration}_x86_64"]["pkg.name-more+"]["cflags"]' in contents
    assert f't_conandeps["{configuration}_x86_64"]["pkg.name-more+"]["sharedlinkflags"]' in contents
    assert f't_conandeps["{configuration}_x86_64"]["pkg.name-more+"]["exelinkflags"]' in contents
    assert f't_conandeps["{configuration}_x86_64"]["pkg.name-more+"]["frameworks"]' in contents
    assert f'if conandeps == nil then conandeps = {{}} end' in contents
    assert f'conan_premake_tmerge(conandeps, t_conandeps)' in contents


def test_premakedeps():
    # Create package
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            name = "pkg.name-more+"
            version = "1.0"

            def package_info(self):
                self.cpp_info.components["libmpdecimal++"].libs = ["libmp++"]
                self.cpp_info.components["mycomp.some-comp+"].libs = ["mylib"]
                self.cpp_info.components["libmpdecimal++"].requires = ["mycomp.some-comp+"]
        """)
    client.save({"conanfile.py": conanfile}, clean_first=True)
    client.run("create . -s arch=x86_64 -s build_type=Debug")
    client.run("create . -s arch=x86_64 -s build_type=Release")

    # Run conan
    client.run("install --require=pkg.name-more+/1.0@ -g PremakeDeps -s arch=x86_64 -s build_type=Debug")
    client.run("install --require=pkg.name-more+/1.0@ -g PremakeDeps -s arch=x86_64 -s build_type=Release")

    # Assert root lua file
    contents = client.load("conandeps.premake5.lua")
    assert 'include "conan_pkg.name-more+.premake5.lua"' in contents
    assert 'function conan_setup_build(conf, pkg)' in contents
    assert 'function conan_setup_link(conf, pkg)' in contents
    assert 'function conan_setup(conf, pkg)' in contents

    # Assert package root file
    contents = client.load("conan_pkg.name-more+.premake5.lua")
    assert 'include "conan_pkg.name-more+_vars_debug_x86_64.premake5.lua"' in contents
    assert 'include "conan_pkg.name-more+_vars_release_x86_64.premake5.lua"' in contents

    # Assert package per configuration files
    assert_vars_file(client, 'debug')
    assert_vars_file(client, 'release')


def test_premakedeps_link_order():
    client = TestClient()
    profile = textwrap.dedent(
        """
        [settings]
        os=Linux
        arch=x86_64
        compiler=gcc
        compiler.version=9
        compiler.cppstd=17
        compiler.libcxx=libstdc++
        build_type=Release
        """)
    client.save(
        {
            "liba/conanfile.py": GenConanfile("liba")
            .with_settings("os", "compiler", "build_type", "arch")
            .with_version("1.0")
            .with_generator("PremakeDeps")
            .with_generator("PremakeToolchain")
            .with_package_info(cpp_info={"libs": ["liba"]}),

            "libb/conanfile.py": GenConanfile("libb")
            .with_settings("os", "compiler", "build_type", "arch")
            .with_version("1.0")
            .with_requires("liba/1.0")
            .with_generator("PremakeDeps")
            .with_package_info(
                cpp_info={
                    "components": {
                        "libb1": {"requires": ["liba::liba"]},
                        "libb2": {"requires": ["libb1"]},
                    }
                }
            ),

            "libc/conanfile.py": GenConanfile("libc")
            .with_settings("os", "compiler", "build_type", "arch")
            .with_version("1.0")
            .with_requires("libb/1.0")
            .with_generator("PremakeDeps")
            .with_package_info(cpp_info={"libs": ["libc"]}),

            "consumer/conanfile.py": GenConanfile("consumer")
            .with_settings("os", "compiler", "build_type", "arch")
            .with_version("1.0")
            .with_requires("libc/1.0")
            .with_generator("PremakeDeps"),

            "profile": profile
        }
    )
    client.run("create liba -pr profile")
    client.run("create libb -pr profile")
    client.run("create libc -pr profile")
    client.run("install consumer -pr profile")
    contents = client.load("consumer/conanconfig.premake5.lua")
    assert 'include "conanconfig_release_x86_64.premake5.lua"' in contents
    contents = client.load("consumer/conanconfig_release_x86_64.premake5.lua")
    # Check correct order of dependencies: more dependent libs should be linked first
    assert 't_conan_deps_order["release_x86_64"] = {"libc", "libb", "liba"}' in contents

    # Check previous configurations are preserved when installing new configuration
    client.run("install consumer -pr profile -s build_type=Debug --build=missing")
    contents = client.load("consumer/conanconfig.premake5.lua")
    assert 'include "conanconfig_release_x86_64.premake5.lua"' in contents
    assert 'include "conanconfig_debug_x86_64.premake5.lua"' in contents
    contents = client.load("consumer/conanconfig_debug_x86_64.premake5.lua")
    assert 't_conan_deps_order["debug_x86_64"] = {"libc", "libb", "liba"}' in contents

@pytest.mark.parametrize("transitive_headers", [True, False])
@pytest.mark.parametrize("transitive_libs", [True, False])
@pytest.mark.parametrize("brotli_package_type", ["unknown", "static-library", "shared-library"])
@pytest.mark.parametrize("lib_package_type", ["unknown", "static-library", "shared-library"])
def test_premakedeps_traits(transitive_headers, transitive_libs, brotli_package_type, lib_package_type):
    client = TestClient()
    profile = textwrap.dedent(
        """
        [settings]
        os=Linux
        arch=x86_64
        compiler=gcc
        compiler.version=9
        compiler.cppstd=17
        compiler.libcxx=libstdc++
        build_type=Release
        """)

    client.save(
        {
            "brotli/conanfile.py": GenConanfile("brotli", "1.0")
            .with_package_type(brotli_package_type)
            .with_package_info(
                {"libs": ["brotlienc2"]}
            ),
            "lib/conanfile.py": GenConanfile("lib", "1.0")
            .with_package_type(lib_package_type)
            .with_requirement(
                "brotli/1.0",
                transitive_headers=transitive_headers,
                transitive_libs=transitive_libs,
            ),
            "conanfile.py": GenConanfile("app", "1.0")
            .with_settings("os", "compiler", "build_type", "arch")
            .with_require("lib/1.0")
            .with_generator("PremakeDeps"),
            "profile": profile,
        }
    )
    client.run("create brotli -pr profile")
    brotli_layout = client.created_layout()
    client.run("create lib -pr profile")
    client.run("install . -pr profile")

    if not transitive_headers and not transitive_libs and brotli_package_type != "shared-library":
        assert not os.path.exists(os.path.join(client.current_folder, "conan_brotli_vars_release_x86_64.premake5.lua"))
        return

    brotli_vars = client.load("conan_brotli_vars_release_x86_64.premake5.lua")
    if transitive_libs:
        assert 't_conandeps["release_x86_64"]["brotli"]["libs"] = {"brotlienc2"}' in brotli_vars
    else:
        assert 't_conandeps["release_x86_64"]["brotli"]["libs"] = {}' in brotli_vars

    if transitive_headers:
        include_path = os.path.join(brotli_layout.package(), "include").replace("\\", "/")
        assert f't_conandeps["release_x86_64"]["brotli"]["includedirs"] = {{"{include_path}"}}' in brotli_vars
    else:
        assert 't_conandeps["release_x86_64"]["brotli"]["includedirs"] = {}' in brotli_vars



import os
import shutil
import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient
from conan.internal.util.files import save


collect = '''\
import os, shutil, fnmatch
from conan.errors import ConanException
from conan.cli.command import conan_command
from conan.api.output import ConanOutput
from conan.cli import make_abs_path
from conan.cli.args import common_graph_args, validate_common_graph_args
from conan.cli.printers import print_profiles
from conan.cli.printers.graph import print_graph_packages, print_graph_basic
# TODO, fix private API
from conan.internal.graph.install_graph import InstallGraph


@conan_command(group="Metadata")
def collect(conan_api, parser, *args):
    """
    command to advanced metadata
    """
    common_graph_args(parser)
    parser.add_argument("-m", "--metadata", action='append',
                        help='Download the metadata matching the pattern, even if the package is '
                             'already in the cache and not downloaded')
    parser.add_argument("-mr", "--metadata-remote", help='Download the metadata from this remote')
    parser.add_argument("-of", "--output-folder", help='The root output folder for metadata')
    parser.add_argument("--build-require", action='store_true', default=False,
                        help='Whether the provided path is a build-require')
    args = parser.parse_args(*args)
    validate_common_graph_args(args)

    # basic paths
    cwd = os.getcwd()
    path = conan_api.local.get_conanfile_path(args.path, cwd, py=None) if args.path else None
    output_folder = (make_abs_path(args.output_folder, cwd) if args.output_folder
                     else os.path.join(cwd, "metadata"))

    # Basic collaborators: remotes, lockfile, profiles
    remotes = conan_api.remotes.list(args.remote) if not args.no_remote else []
    lockfile = conan_api.lockfile.get_lockfile(lockfile=args.lockfile, conanfile_path=path, cwd=cwd,
                                               partial=args.lockfile_partial)
    profile_host, profile_build = conan_api.profiles.get_profiles_from_args(args)
    print_profiles(profile_host, profile_build)

    # Graph computation (without installation of binaries)
    gapi = conan_api.graph
    if path:
        deps_graph = gapi.load_graph_consumer(path, args.name, args.version, args.user, args.channel,
                                              profile_host, profile_build, lockfile, remotes,
                                              args.update, is_build_require=args.build_require)
    else:
        deps_graph = gapi.load_graph_requires(args.requires, args.tool_requires, profile_host,
                                              profile_build, lockfile, remotes, args.update)
    print_graph_basic(deps_graph)
    deps_graph.report_graph_error()
    gapi.analyze_binaries(deps_graph, args.build, remotes, update=args.update, lockfile=lockfile)
    print_graph_packages(deps_graph)

    # Installation of binaries and consumer generators
    conan_api.install.install_binaries(deps_graph=deps_graph, remotes=remotes)

    # Downloading metadata
    metadata = args.metadata if args.metadata else ["*"]
    # TODO: This InstallGraph API is EXPERIMENTAL and will change in future versions
    install_graph = InstallGraph(deps_graph)
    install_order = install_graph.install_order(flat=True)

    if args.metadata_remote:
        out = ConanOutput()
        out.title(f"Downloading metadata from {args.metadata_remote}")
        remote = conan_api.remotes.get(args.metadata_remote)

        for install_reference in install_order:
            if install_reference.ref.revision is None:  # Is an editable/system, do not download
                continue
            try:
                conan_api.download.recipe(install_reference.ref, remote, metadata)
            except ConanException as e:
                out.warning(f"Recipe {install_reference.ref} not found in remote: {e}")
                continue
            for package in install_reference.packages.values():
                try:
                    conan_api.download.package(package.pref, remote, metadata)
                except ConanException as e:
                    out.warning(f"Package {package.pref} not found in remote: {e}")

    # Copying and collecting metadata from all packages into local copy
    def _copy_metadata(src, dst):
        for root, subfolders, files in os.walk(src):
            relative_path = os.path.relpath(root, src)
            for f in files:
                relative_name = os.path.normpath(os.path.join(relative_path, f))
                if any(fnmatch.fnmatch(relative_name, m) for m in metadata):
                    os.makedirs(os.path.dirname(os.path.join(dst, relative_name)), exist_ok=True)
                    shutil.copy2(os.path.join(src, relative_name), os.path.join(dst, relative_name))

    for install_reference in install_order:
        conanfile = install_reference.node.conanfile
        if conanfile.recipe_metadata_folder is None:
            continue
        folder = os.path.join(output_folder, conanfile.ref.name, str(conanfile.ref.version))
        if os.path.exists(folder):
            conanfile.output.warning(f"Folder for {conanfile} already exist, removing it: {folder}")
            shutil.rmtree(folder)
        conanfile.output.info(f"Copying recipe metadata from {conanfile.recipe_metadata_folder}")
        _copy_metadata(conanfile.recipe_metadata_folder, os.path.join(folder, "recipe"))
        for package in install_reference.packages.values():
            pkg_metadata_folder = package.conanfile.package_metadata_folder
            conanfile.output.info(f"Copying package metadata from {pkg_metadata_folder}")
            _copy_metadata(pkg_metadata_folder, os.path.join(folder, "package"))
'''


def test_custom_command_collect_no_metadata():
    c = TestClient(default_server_user=True)
    command_file_path = os.path.join(c.cache_folder, 'extensions',
                                     'commands', 'metadata', 'cmd_collect.py')
    save(command_file_path, collect)
    c.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
            "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_requires("dep/0.1")})
    c.run("create dep")
    c.run("create pkg")
    c.run("metadata:collect --requires=pkg/0.1 --metadata=* --metadata-remote=default")
    # It does nothing, but it doesn't crash

    c.run("upload * -r=default -c")
    c.run("metadata:collect --requires=pkg/0.1 --metadata=* --metadata-remote=default")
    # It does nothing, but it doesn't crash

    c.run("editable add dep")
    c.run("metadata:collect --requires=pkg/0.1 --metadata=* --metadata-remote=default")
    # It does nothing, but it doesn't crash


conanfile = textwrap.dedent("""
    import os
    from conan import ConanFile
    from conan.tools.files import save, copy

    class Pkg(ConanFile):
        name = "{name}"
        version = "0.1"
        {requires}

        def layout(self):
            self.folders.build = "build"

        def source(self):
            save(self, os.path.join(self.recipe_metadata_folder, "logs", "src.log"),
                 f"srclog {{self.name}}!!")

        def build(self):
            save(self, "mylogs.txt", f"some logs {{self.name}}!!!")
            copy(self, "mylogs.txt", src=self.build_folder,
                 dst=os.path.join(self.package_metadata_folder, "logs"))
    """)


def test_custom_command_collect():
    c = TestClient(default_server_user=True)
    command_file_path = os.path.join(c.cache_folder, 'extensions',
                                     'commands', 'metadata', 'cmd_collect.py')
    save(command_file_path, collect)
    c.save({"dep/conanfile.py": conanfile.format(name="dep",
                                                 requires="tool_requires = 'tool/0.1'"),
            "pkg/conanfile.py": conanfile.format(name="pkg", requires='requires = "dep/0.1"'),
            "profile": "[platform_tool_requires]\ntool/0.1"})
    c.run("create dep -pr=profile")
    c.run("create pkg -pr=profile")
    c.run("metadata:collect --requires=pkg/0.1 --metadata=* --metadata-remote=default -pr=profile")
    assert "srclog dep!!" in c.load("metadata/dep/0.1/recipe/logs/src.log")
    assert "some logs dep!!!" in c.load("metadata/dep/0.1/package/logs/mylogs.txt")
    assert "srclog pkg!!" in c.load("metadata/pkg/0.1/recipe/logs/src.log")
    assert "some logs pkg!!!" in c.load("metadata/pkg/0.1/package/logs/mylogs.txt")

    shutil.rmtree(os.path.join(c.current_folder, "metadata"))

    c.run("upload * -r=default -c")
    c.run("remove * -c")
    c.run("metadata:collect --requires=pkg/0.1 --metadata=* --metadata-remote=default -pr=profile")
    assert "srclog dep!!" in c.load("metadata/dep/0.1/recipe/logs/src.log")
    assert "some logs dep!!!" in c.load("metadata/dep/0.1/package/logs/mylogs.txt")
    assert "srclog pkg!!" in c.load("metadata/pkg/0.1/recipe/logs/src.log")
    assert "some logs pkg!!!" in c.load("metadata/pkg/0.1/package/logs/mylogs.txt")

    shutil.rmtree(os.path.join(c.current_folder, "metadata"))
    c.run("editable add dep")
    c.run("source dep")
    c.run("build dep -pr=profile")
    c.run("metadata:collect --requires=pkg/0.1 --metadata=* --metadata-remote=default -pr=profile")
    assert "srclog dep!!" in c.load("metadata/dep/0.1/recipe/logs/src.log")
    assert "some logs dep!!!" in c.load("metadata/dep/0.1/package/logs/mylogs.txt")
    assert "srclog pkg!!" in c.load("metadata/pkg/0.1/recipe/logs/src.log")
    assert "some logs pkg!!!" in c.load("metadata/pkg/0.1/package/logs/mylogs.txt")



from conan.test.utils.tools import GenConanfile, TestClient


def test_basic_parallel_download():
    client = TestClient(light=True, default_server_user=True)
    threads = 2
    packages = 2
    per_package = 4
    client.save_home({"global.conf": f"core.download:parallel={threads}"})
    client.save({"conanfile.py": GenConanfile().with_option("myoption", '["ANY"]')})

    package_ids = []
    for i in range(packages):
        for n in range(per_package):
            client.run(f"create . --name=pkg{i} --version=0.1 --user=user --channel=testing -o pkg{i}/*:myoption={n}")
            package_id = client.created_package_id(f"pkg{i}/0.1@user/testing")
            package_ids.append((i, package_id))
    client.run("upload * --confirm -r default")
    client.run("remove * -c")

    # Lets download the packages
    client.run("download pkg*/0.1@user/testing#*:* -r default")
    assert f"Downloading with {threads} parallel threads" in client.out
    for i, package_id in package_ids:
        assert f"pkg{i}/0.1@user/testing: Package installed {package_id}" in client.out



import platform
import textwrap

import pytest

from conan.test.utils.tools import TestClient


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
@pytest.mark.tool("visual_studio", "17")
def test_deactivate_vcvars_message():
    client = TestClient()
    conanfile = textwrap.dedent("""
            from conan import ConanFile
            class TestConan(ConanFile):
                generators = "VCVars"
                settings = "os", "compiler", "arch", "build_type"
        """)
    client.save({"conanfile.py": conanfile})
    client.run('install . -s compiler.version=194')
    client.run_command(r'conanbuild.bat')
    assert "[vcvarsall.bat] Environment initialized" in client.out
    client.run_command(r'deactivate_conanvcvars.bat')
    assert "vcvars env cannot be deactivated" in client.out


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows Powershell")
def test_deactivate_vcvars_with_powershell():
    client = TestClient()
    conanfile = textwrap.dedent("""
                from conan import ConanFile
                class TestConan(ConanFile):
                    generators = "VCVars"
                    settings = "os", "compiler", "arch", "build_type"
            """)
    client.save({"conanfile.py": conanfile})
    client.run('install . -c tools.env.virtualenv:powershell=True')
    client.run_command(r'powershell.exe ".\conanbuild.ps1"')
    assert "conanvcvars.ps1: Activated environment" in client.out
    client.run_command(r'powershell.exe ".\deactivate_conanvcvars.ps1"')
    assert "vcvars env cannot be deactivated" in client.out



import os
import platform

import pytest

from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient

import textwrap

_path_chars = "päthñç$"


@pytest.fixture(scope="module")
def client_with_special_chars():
    """ the path with special characters is creating a conanbuild.bat that fails
    """
    cache_folder = os.path.join(temp_folder(), _path_chars)
    current_folder = os.path.join(temp_folder(), _path_chars)
    c = TestClient(cache_folder, current_folder)

    tool = textwrap.dedent(r"""
        import os
        from conan import ConanFile
        from conan.tools.files import save, chdir
        class Pkg(ConanFile):
            name = "mytool"
            version = "1.0"
            def package(self):
                with chdir(self, self.package_folder):
                    echo = "@echo off\necho MYTOOL WORKS!!"
                    save(self, "bin/mytool.bat", echo)
                    save(self, "bin/mytool.sh", echo)
                    os.chmod("bin/mytool.sh", 0o777)
            """)
    c.save({"conanfile.py": tool})
    c.run("create .")

    conan_file = textwrap.dedent("""
        import platform
        from conan import ConanFile

        class App(ConanFile):
            name="failure"
            version="0.1"
            settings = 'os', 'arch', 'compiler', 'build_type'
            generators = "VirtualBuildEnv"
            tool_requires = "mytool/1.0"
            apply_env = False # SUPER IMPORTANT, DO NOT REMOVE

            def build(self):
                mycmd = "mytool.bat" if platform.system() == "Windows" else "mytool.sh"
                self.run(mycmd)
        """)
    c.save({"conanfile.py": conan_file})
    return c


def test_reuse_buildenv(client_with_special_chars):
    c = client_with_special_chars
    # Need the 2 profile to work correctly buildenv
    c.run("create .")
    assert _path_chars in c.out
    assert "MYTOOL WORKS!!" in c.out


@pytest.mark.skipif(platform.system() != "Windows", reason="powershell only win")
def test_reuse_buildenv_powershell(client_with_special_chars):
    c = client_with_special_chars
    c.run("create . -c tools.env.virtualenv:powershell=True")
    assert _path_chars in c.out
    assert "MYTOOL WORKS!!" in c.out



import json
from collections import OrderedDict

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.env import environment_update
from conan.test.utils.tools import TestClient, TestServer


class TestListUpload:
    refs = ["zli/1.0.0#f034dc90894493961d92dd32a9ee3b78",
            "zlib/1.0.0@user/channel#ffd4bc45820ddb320ab224685b9ba3fb"]

    @pytest.fixture()
    def client(self):
        c = TestClient(default_server_user=True, light=True)
        c.save({
            "zlib.py": GenConanfile("zlib"),
            "zli.py": GenConanfile("zli", "1.0.0")
        })
        c.run("create zli.py")
        c.run("create zlib.py --version=1.0.0 --user=user --channel=channel")
        return c

    def test_list_upload_recipes(self, client):
        pattern = "z*#latest"
        client.run(f"list {pattern} --format=json", redirect_stdout="pkglist.json")
        client.run("upload --list=pkglist.json -r=default")
        for r in self.refs:
            assert f"Uploading recipe '{r}'" in client.out
        assert "Uploading package" not in client.out

    def test_list_upload_packages(self, client):
        pattern = "z*#latest:*#latest"
        client.run(f"list {pattern} --format=json", redirect_stdout="pkglist.json")
        client.run("upload --list=pkglist.json -r=default")
        for r in self.refs:
            assert f"Uploading recipe '{r}'" in client.out
        assert str(client.out).count("Uploading package") == 2

    def test_list_upload_empty_list(self, client):
        client.run(f"install --requires=zlib/1.0.0@user/channel -f json",
                   redirect_stdout="install_graph.json")

        # Generate an empty pkglist.json
        client.run(f"list --format=json --graph=install_graph.json --graph-binaries=bogus",
                   redirect_stdout="pkglist.json")

        # No binaries should be uploaded since the pkglist is empty, but the command
        # should not error
        client.run("upload --list=pkglist.json -r=default")
        assert "No packages were uploaded because the package list is empty." in client.out


class TestGraphCreatedUpload:
    def test_create_upload(self):
        c = TestClient(default_server_user=True, light=True)
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0"),
                "app/conanfile.py": GenConanfile("app", "1.0").with_requires("zlib/1.0")})
        c.run("create zlib")
        c.run("create app --format=json", redirect_stdout="graph.json")
        c.run("list --graph=graph.json --format=json", redirect_stdout="pkglist.json")
        c.run("upload --list=pkglist.json -r=default")
        assert "Uploading recipe 'app/1.0#0fa1ff1b90576bb782600e56df642e19'" in c.out
        assert "Uploading recipe 'zlib/1.0#c570d63921c5f2070567da4bf64ff261'" in c.out
        assert "Uploading package 'app" in c.out
        assert "Uploading package 'zlib" in c.out


class TestExportUpload:
    def test_export_upload(self):
        c = TestClient(default_server_user=True, light=True)
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0")})
        c.run("export zlib --format=pkglist", redirect_stdout="pkglist.json")
        c.run("upload --list=pkglist.json -r=default -c")
        assert "Uploading recipe 'zlib/1.0#c570d63921c5f2070567da4bf64ff261'" in c.out


class TestCreateGraphToPkgList:
    def test_graph_pkg_nonexistant(self):
        c = TestClient(light=True)
        c.run("list --graph=non-existent-file.json", assert_error=True)
        assert "ERROR: Graph file not found" in c.out

    def test_graph_pkg_list_only_built(self):
        c = TestClient(light=True)
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0"),
                "app/conanfile.py": GenConanfile("app", "1.0").with_requires("zlib/1.0")
                                                              .with_settings("os")
                                                              .with_shared_option(False)})
        c.run("create zlib")
        c.run("create app --format=json -s os=Linux", redirect_stdout="graph.json")
        c.run("list --graph=graph.json --graph-binaries=build --format=json")
        pkglist = json.loads(c.stdout)["Local Cache"]
        assert len(pkglist) == 1
        pkgs = pkglist["app/1.0"]["revisions"]["8263c3c32802e14a2f03a0b1fcce0d95"]["packages"]
        assert len(pkgs) == 1
        pkg_app = pkgs["d8b3bdd894c3eb9bf2a3119ee0f8c70843ace0ac"]
        assert pkg_app["info"]["requires"] == ["zlib/1.0.Z"]
        assert pkg_app["info"]["settings"] == {'os': 'Linux'}
        assert pkg_app["info"]["options"] == {'shared': 'False'}

    def test_graph_pkg_list_all_recipes_only(self):
        """
        --graph-recipes=* selects all the recipes in the graph
        """
        c = TestClient(light=True)
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0"),
                "app/conanfile.py": GenConanfile("app", "1.0").with_requires("zlib/1.0")})
        c.run("create zlib")
        c.run("create app --format=json", redirect_stdout="graph.json")
        c.run("list --graph=graph.json --graph-recipes=* --format=json")
        pkglist = json.loads(c.stdout)["Local Cache"]
        assert len(pkglist) == 2
        assert "packages" not in pkglist["app/1.0"]["revisions"]["0fa1ff1b90576bb782600e56df642e19"]
        assert "packages" not in pkglist["zlib/1.0"]["revisions"]["c570d63921c5f2070567da4bf64ff261"]

    def test_graph_pkg_list_python_requires(self):
        """
        include python_requires too
        """
        c = TestClient(default_server_user=True, light=True)
        c.save({"pytool/conanfile.py": GenConanfile("pytool", "0.1"),
                "zlib/conanfile.py": GenConanfile("zlib", "1.0").with_python_requires("pytool/0.1"),
                "app/conanfile.py": GenConanfile("app", "1.0").with_requires("zlib/1.0")})
        c.run("create pytool")
        c.run("create zlib")
        c.run("upload * -c -r=default")
        c.run("remove * -c")
        c.run("create app --format=json", redirect_stdout="graph.json")
        c.run("list --graph=graph.json --format=json")
        pkglist = json.loads(c.stdout)["Local Cache"]
        assert len(pkglist) == 3
        assert "96aec08148a2392127462c800e1c8af6" in pkglist["pytool/0.1"]["revisions"]
        pkglist = json.loads(c.stdout)["default"]
        assert len(pkglist) == 2
        assert "96aec08148a2392127462c800e1c8af6" in pkglist["pytool/0.1"]["revisions"]

    def test_graph_pkg_list_create_python_requires(self):
        """
        include python_requires too
        """
        c = TestClient(default_server_user=True, light=True)
        c.save({"conanfile.py": GenConanfile("pytool", "0.1").with_package_type("python-require")})
        c.run("create . --format=json", redirect_stdout="graph.json")
        c.run("list --graph=graph.json --format=json")
        pkglist = json.loads(c.stdout)["Local Cache"]
        assert len(pkglist) == 1
        assert "62a6a9e5347b789bfc6572948ea19f85" in pkglist["pytool/0.1"]["revisions"]


class TestGraphInfoToPkgList:
    def test_graph_pkg_list_only_built(self):
        c = TestClient(default_server_user=True, light=True)
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0"),
                "app/conanfile.py": GenConanfile("app", "1.0").with_requires("zlib/1.0")})
        c.run("create zlib")
        c.run("create app --format=json")
        c.run("upload * -c -r=default")
        c.run("remove * -c")
        c.run("graph info --requires=app/1.0 --format=json", redirect_stdout="graph.json")
        c.run("list --graph=graph.json --graph-binaries=build --format=json")
        pkglist = json.loads(c.stdout)
        assert len(pkglist["Local Cache"]) == 0
        assert len(pkglist["default"]) == 2
        c.run("install --requires=app/1.0 --format=json", redirect_stdout="graph.json")
        c.run("list --graph=graph.json --graph-binaries=download --format=json")
        pkglist = json.loads(c.stdout)
        assert len(pkglist["Local Cache"]) == 2
        assert len(pkglist["default"]) == 2
        c.run("list --graph=graph.json --graph-binaries=build --format=json")
        pkglist = json.loads(c.stdout)
        assert len(pkglist["Local Cache"]) == 0
        assert len(pkglist["default"]) == 2


class TestPkgListFindRemote:
    """ we can recover a list of remotes for an already installed graph, for metadata download
    """
    def test_graph_2_pkg_list_remotes(self):
        servers = OrderedDict([("default", TestServer()), ("remote2", TestServer())])
        c = TestClient(servers=servers, inputs=2 * ["admin", "password"], light=True)
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0"),
                "app/conanfile.py": GenConanfile("app", "1.0").with_requires("zlib/1.0")})
        c.run("create zlib")
        c.run("create app ")
        c.run("upload zlib* -c -r=default")
        c.run("upload zlib* -c -r=remote2")
        c.run("upload app* -c -r=remote2")

        # This install, packages will be in the cache
        c.run("install --requires=app/1.0 --format=json", redirect_stdout="graph.json")
        # So list, will not have remote at all
        c.run("list --graph=graph.json --format=json", redirect_stdout="pkglist.json")

        pkglist = json.loads(c.load("pkglist.json"))
        assert len(pkglist["Local Cache"]) == 2
        assert "default" not in pkglist  # The remote doesn't even exist

        # Lets now compute a list finding in the remotes
        c.run("pkglist find-remote pkglist.json --format=json", redirect_stdout="remotepkg.json")
        pkglist = json.loads(c.stdout)
        assert "Local Cache" not in pkglist
        assert len(pkglist["default"]) == 1
        assert "zlib/1.0" in pkglist["default"]
        assert len(pkglist["remote2"]) == 2
        assert "app/1.0" in pkglist["remote2"]
        assert "zlib/1.0" in pkglist["remote2"]

        c.run("download --list=remotepkg.json -r=default --metadata=*")
        assert "zlib/1.0: Retrieving recipe metadata from remote 'default'" in c.out
        assert "zlib/1.0: Retrieving package metadata" in c.out
        c.run("download --list=remotepkg.json -r=remote2 --metadata=*")
        assert "app/1.0: Retrieving recipe metadata from remote 'remote2'" in c.out
        assert "app/1.0: Retrieving package metadata" in c.out

    def test_input_only_name_version(self):
        c = TestClient(default_server_user=True, light=True)
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0")})
        c.run("create zlib")
        c.run("upload zlib* -c -r=default")

        # Create input pkglist for find-remote
        c.run(f"list * --format=json", redirect_stdout="mylist.json")
        pkglist = json.loads(c.load("mylist.json"))
        expected = {"zlib/1.0": {}}
        assert pkglist["Local Cache"] == expected

        c.run("pkglist find-remote mylist.json --format=json --remote default")
        pkglist = json.loads(c.stdout)
        assert pkglist["default"] == expected

    def test_input_recipe_revisions(self):
        c = TestClient(default_server_user=True, light=True)
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0")})
        c.run("create zlib")
        c.run("upload zlib* -c -r=default")

        # Create input pkglist for find-remote
        c.run(f"list *#* --format=json", redirect_stdout="mylist.json")

        def _check(origin):
            pkglist = json.loads(c.load("mylist.json"))
            revs = pkglist[origin]["zlib/1.0"]["revisions"]
            assert list(revs) == ["c570d63921c5f2070567da4bf64ff261"]
            assert "packages" not in revs["c570d63921c5f2070567da4bf64ff261"]

        _check("Local Cache")

        c.run("pkglist find-remote mylist.json --format=json --remote default",
              redirect_stdout="mylist.json")
        _check("default")

    def test_input_only_package_ids(self):
        c = TestClient(default_server_user=True, light=True)
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0").with_settings("os")})
        c.run("create zlib -s os=Linux")
        c.run("upload zlib* -c -r=default")

        # Create input pkglist for find-remote
        c.run(f"list *:* --format=json", redirect_stdout="mylist.json")

        def _check(origin):
            pkglist = json.loads(c.load("mylist.json"))
            revs = pkglist[origin]["zlib/1.0"]["revisions"]
            assert list(revs) == ["1cb7410d0365f87510a6767c7bef804e"]
            info = {"info": {'settings': {'os': 'Linux'}}}
            expected = {"9a4eb3c8701508aa9458b1a73d0633783ecc2270": info}
            assert revs["1cb7410d0365f87510a6767c7bef804e"]["packages"] == expected

        c.run("pkglist find-remote mylist.json --format=json --remote default",
              redirect_stdout="mylist.json")
        _check("default")

    def test_graph_pkg_list_of_recipes_and_binaries(self):
        c = TestClient(default_server_user=True, light=True)
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0").with_settings("os")})
        c.run("create zlib -s os=Linux")
        c.run("upload zlib* -c -r=default")

        # Create input pkglist for find-remote
        c.run(f"list *#*:*#* --format=json", redirect_stdout="mylist.json")

        def _check(origin):
            pkglist = json.loads(c.load("mylist.json"))
            revs = pkglist[origin]["zlib/1.0"]["revisions"]
            assert list(revs) == ["1cb7410d0365f87510a6767c7bef804e"]
            expected = {'settings': {'os': 'Linux'}}
            pkgs = revs["1cb7410d0365f87510a6767c7bef804e"]["packages"]
            assert list(pkgs) == ["9a4eb3c8701508aa9458b1a73d0633783ecc2270"]
            pkg = pkgs["9a4eb3c8701508aa9458b1a73d0633783ecc2270"]
            assert pkg["info"] == expected
            assert list(pkg["revisions"]) == ["1d3c57385f4133c1fbd6d13bd538496e"]

        _check("Local Cache")
        c.run("pkglist find-remote mylist.json --format=json --remote default",
              redirect_stdout="mylist.json")
        _check("default")

    def test_graph_pkg_list_counter_example(self):
        c = TestClient(default_server_user=True, light=True)
        c.save({"conanfile.py": GenConanfile("zlib", "1.0").with_package_file("file.txt",
                                                                              env_var="MY_VAR")})

        with environment_update({"MY_VAR": "1"}):
            c.run("create .")
        with environment_update({"MY_VAR": "2"}):
            c.run("create .")
        c.run("upload zlib*:*#* -c -r=default")

        # Create input pkglist for find-remote
        c.run(f"list zlib/1.0#latest:*#latest --format=json", redirect_stdout="mylist.json")

        def _check(origin):
            pkglist = json.loads(c.load("mylist.json"))
            prevs = pkglist[origin]["zlib/1.0"]["revisions"]
            input_pkgs = prevs["212b9babae6a4b8a8362703cec4257ad"]["packages"]
            prevs = input_pkgs["da39a3ee5e6b4b0d3255bfef95601890afd80709"]["revisions"]
            assert len(prevs) == 1

        _check("Local Cache")
        c.run("pkglist find-remote mylist.json --format=json --remote default",
              redirect_stdout="mylist.json")
        _check("default")


class TestPkgListMerge:
    """ deep merge lists
    """
    def test_graph_2_pkg_list_remotes(self):
        servers = OrderedDict([("default", TestServer()), ("remote2", TestServer())])
        c = TestClient(servers=servers, inputs=2 * ["admin", "password"])
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0").with_settings("build_type"),
                "bzip2/conanfile.py": GenConanfile("bzip2", "1.0").with_settings("build_type"),
                "app/conanfile.py": GenConanfile("app", "1.0").with_requires("zlib/1.0", "bzip2/1.0")
                                                              .with_settings("build_type")})
        c.run("create zlib")
        c.run("create bzip2")
        c.run("create app ")

        c.run("list zlib:* --format=json", redirect_stdout="list1.json")
        c.run("list bzip2:* --format=json", redirect_stdout="list2.json")
        c.run("list app:* --format=json", redirect_stdout="list3.json")
        c.run("pkglist merge --list=list1.json --list=list2.json --list=list3.json --format=json",
              redirect_stdout="release.json")
        final = json.loads(c.stdout)
        assert "app/1.0" in final["Local Cache"]
        assert "zlib/1.0" in final["Local Cache"]
        assert "bzip2/1.0" in final["Local Cache"]

        c.run("create zlib -s build_type=Debug")
        c.run("create bzip2 -s build_type=Debug")
        c.run("create app -s build_type=Debug")
        c.run("list *:* -fs build_type=Debug --format=json", redirect_stdout="debug.json")
        c.run("pkglist merge --list=release.json --list=debug.json --format=json",
              redirect_stdout="release.json")
        final = json.loads(c.stdout)
        rev = final["Local Cache"]["zlib/1.0"]["revisions"]["11f74ff5f006943c6945117511ac8b64"]
        assert len(rev["packages"]) == 2  # Debug and Release
        settings = rev["packages"]["efa83b160a55b033c4ea706ddb980cd708e3ba1b"]["info"]["settings"]
        assert settings == {"build_type": "Release"}
        settings = rev["packages"]["9e186f6d94c008b544af1569d1a6368d8339efc5"]["info"]["settings"]
        assert settings == {"build_type": "Debug"}
        rev = final["Local Cache"]["bzip2/1.0"]["revisions"]["9e0352b3eb99ba4ac79bc7eeae2102c5"]
        assert len(rev["packages"]) == 2  # Debug and Release
        settings = rev["packages"]["efa83b160a55b033c4ea706ddb980cd708e3ba1b"]["info"]["settings"]
        assert settings == {"build_type": "Release"}
        settings = rev["packages"]["9e186f6d94c008b544af1569d1a6368d8339efc5"]["info"]["settings"]
        assert settings == {"build_type": "Debug"}

    def test_pkglist_file_error(self):
        # This can happen when reusing the same file in input and output
        c = TestClient(light=True)
        c.run("pkglist merge -l mylist.json", assert_error=True)
        assert "ERROR: Package list file missing or broken:" in c.out
        c.save({"mylist.json": ""})
        c.run("pkglist merge -l mylist.json", assert_error=True)
        assert "ERROR: Package list file invalid JSON:" in c.out

        # This can happen when using a graph JSON file instead of a package list file
        c.save({"conanfile.py": GenConanfile("lib")})
        c.run("create . --version=1.0 --format=json", redirect_stdout="out/graph.json")
        c.run("pkglist merge -l out/graph.json", assert_error=True)
        assert (
            'ERROR: Expected a package list file but found a graph file. '
            'You can create a "package list" JSON file by running:' in c.out
        )
        assert "conan list --graph graph.json --format=json > pkglist.json" in c.out


class TestDownloadUpload:
    @pytest.fixture()
    def client(self):
        c = TestClient(default_server_user=True, light=True)
        c.save({
            "zlib.py": GenConanfile("zlib"),
            "zli.py": GenConanfile("zli", "1.0.0")
        })
        c.run("create zli.py")
        c.run("create zlib.py --version=1.0.0 --user=user --channel=channel")
        c.run("upload * -r=default -c")
        c.run("remove * -c")
        return c

    @pytest.mark.parametrize("prev_list", [False, True])
    def test_download_upload_all(self, client, prev_list):
        # We need to be consequeent with the pattern, it is not the same defaults for
        # download and for list
        pattern = "zlib/*#latest:*#latest"
        if prev_list:
            client.run(f"list {pattern} -r=default --format=json", redirect_stdout="pkglist.json")
            # Overwriting previous pkglist.json
            pattern = "--list=pkglist.json"

        client.run(f"download {pattern} -r=default --format=json", redirect_stdout="pkglist.json")
        # TODO: Discuss "origin"
        assert "Local Cache" in client.load("pkglist.json")
        client.run("remove * -r=default -c")
        client.run("upload --list=pkglist.json -r=default")
        assert f"Uploading recipe 'zlib/1.0.0" in client.out
        assert f"Uploading recipe 'zli/" not in client.out
        assert "Uploading package 'zlib/1.0.0" in client.out
        assert "Uploading package 'zli/" not in client.out

    def test_download_upload_all_no_removed(self):
        c = TestClient(default_server_user=True, light=True)
        c.save({"zlib.py": GenConanfile("zlib", "0.1")})
        c.run("create zlib.py ")
        c.run("upload * -r=default -c --format=json", redirect_stdout="pkglist.json")
        c.run("remove * -c")
        c.run(f"download --list=pkglist.json -r=default --format=json",
              redirect_stdout="pkglist.json")
        # The original "files" and "upload-urls" fields of first upload are removed
        pkglist_json = c.load("pkglist.json")
        assert "files" not in pkglist_json
        assert "upload-urls" not in pkglist_json
        c.run("upload --list=pkglist.json -r=default --format=json")
        # It doesn't crash anymore

    @pytest.mark.parametrize("prev_list", [False, True])
    def test_download_upload_only_recipes(self, client, prev_list):
        if prev_list:
            pattern = "zlib/*#latest"
            client.run(f"list {pattern} -r=default --format=json", redirect_stdout="pkglist.json")
            # Overwriting previous pkglist.json
            pattern = "--list=pkglist.json"
        else:
            pattern = "zlib/*#latest --only-recipe"
        client.run(f"download {pattern} -r=default --format=json", redirect_stdout="pkglist.json")
        # TODO: Discuss "origin"
        assert "Local Cache" in client.load("pkglist.json")
        # Download binary too! Just to make sure it is in the cache, but not uploaded
        # because it is not in the orignal list of only recipes
        client.run(f"download * -r=default")
        client.run("remove * -r=default -c")
        client.run("upload --list=pkglist.json -r=default")
        assert f"Uploading recipe 'zlib/1.0.0" in client.out
        assert f"Uploading recipe 'zli/" not in client.out
        assert "Uploading package 'zlib/1.0.0" not in client.out
        assert "Uploading package 'zli/" not in client.out


class TestListRemove:
    @pytest.fixture()
    def client(self):
        c = TestClient(default_server_user=True, light=True)
        c.save({
            "zlib.py": GenConanfile("zlib"),
            "zli.py": GenConanfile("zli", "1.0.0")
        })
        c.run("create zli.py")
        c.run("create zlib.py --version=1.0.0 --user=user --channel=channel")
        c.run("upload * -r=default -c")
        return c

    def test_remove_nothing_only_refs(self, client):
        # It is necessary to do *#* for actually removing something
        client.run(f"list * --format=json", redirect_stdout="pkglist.json")
        client.run(f"remove --list=pkglist.json -c --format=json")
        assert "Nothing to remove, package list do not contain recipe revisions" in client.out
        result = json.loads(client.stdout)
        assert result["Local Cache"] == {}  # Nothing was removed

    @pytest.mark.parametrize("remote", [False, True])
    def test_remove_all(self, client, remote):
        # It is necessary to do *#* for actually removing something
        remote = "-r=default" if remote else ""
        client.run(f"list *#* {remote} --format=json", redirect_stdout="pkglist.json")
        client.run(f"remove --list=pkglist.json {remote} -c --dry-run")
        assert "zli/1.0.0#f034dc90894493961d92dd32a9ee3b78:" \
               " Removed recipe and all binaries" in client.out
        assert "zlib/1.0.0@user/channel#ffd4bc45820ddb320ab224685b9ba3fb:" \
               " Removed recipe and all binaries" in client.out

        client.run(f"remove --list=pkglist.json {remote} -c --format=json")
        result = json.loads(client.stdout)
        origin = "Local Cache" if not remote else "default"
        assert len(result[origin]["zli/1.0.0"]["revisions"]) == 1
        assert len(result[origin]["zlib/1.0.0@user/channel"]["revisions"]) == 1
        assert "packages" not in client.stdout  # Packages are not listed at all

        client.run(f"list * {remote}")
        assert "There are no matching recipe references" in client.out

    @pytest.mark.parametrize("remote", [False, True])
    def test_remove_packages_no_revisions(self, client, remote):
        # It is necessary to do *#*:*#* for actually removing binaries
        remote = "-r=default" if remote else ""
        client.run(f"list *#*:* {remote} --format=json", redirect_stdout="pkglist.json")
        client.run(f"remove --list=pkglist.json {remote} -c --format=json")
        assert "No binaries to remove for 'zli/1.0.0#f034dc90894493961d92dd32a9ee3b78'" in client.out
        assert "No binaries to remove for 'zlib/1.0.0@user/channel" \
               "#ffd4bc45820ddb320ab224685b9ba3fb" in client.out
        result = json.loads(client.stdout)
        origin = "Local Cache" if not remote else "default"
        assert result[origin] == {}  # Nothing was removed

    @pytest.mark.parametrize("remote", [False, True])
    def test_remove_packages(self, client, remote):
        # It is necessary to do *#*:*#* for actually removing binaries
        remote = "-r=default" if remote else ""
        client.run(f"list *#*:*#* {remote} --format=json", redirect_stdout="pkglist.json")
        client.run(f"remove --list=pkglist.json {remote} -c --dry-run")
        assert "Removed recipe and all binaries" not in client.out
        assert "zli/1.0.0#f034dc90894493961d92dd32a9ee3b78: Removed binaries" in client.out
        assert "zlib/1.0.0@user/channel#ffd4bc45820ddb320ab224685b9ba3fb: " \
               "Removed binaries" in client.out

        client.run(f"remove --list=pkglist.json {remote} -c --format=json")
        result = json.loads(client.stdout)
        origin = "Local Cache" if not remote else "default"
        zli_revs = result[origin]["zli/1.0.0"]["revisions"]
        zli_uc_revs = result[origin]["zlib/1.0.0@user/channel"]["revisions"]
        assert len(zli_revs) == 1
        assert len(zli_uc_revs) == 1
        assert len(zli_revs["f034dc90894493961d92dd32a9ee3b78"]["packages"]) == 1
        assert len(zli_uc_revs["ffd4bc45820ddb320ab224685b9ba3fb"]["packages"]) == 1

        client.run(f"list *:* {remote}")
        assert "zli/1.0.0" in client.out
        assert "zlib/1.0.0@user/channel" in client.out


class TestListGraphContext:

    @pytest.mark.parametrize("context, refs, not_refs", [
        ("build", ["cmake/1.0", "m4/1.0", "protobuf/1.0"], []),
        ("build-only", ["cmake/1.0", "m4/1.0"], ["protobuf/1.0"]),
        ("host", ["zlib/1.0", "protobuf/1.0"], []),
        ("host-only", ["zlib/1.0"], ["protobuf/1.0"])
    ])
    def test_list_graph_context(self, context, refs, not_refs):
        c = TestClient(default_server_user=True, light=True)
        c.save({
            "zlib/conanfile.py": GenConanfile("zlib", "1.0").with_settings("os"),
            "cmake/conanfile.py": GenConanfile("cmake", "1.0").with_settings("os"),
            "m4/conanfile.py": GenConanfile("m4", "1.0").with_settings("os"),
            "protobuf/conanfile.py": GenConanfile("protobuf", "1.0").with_settings("os"),
            "app/conanfile.py": GenConanfile("app", "1.0").with_settings("os")
            .with_requires("zlib/1.0", "protobuf/1.0")
            .with_tool_requires("cmake/1.0", "m4/1.0", "protobuf/1.0")})
        c.run("create zlib")
        c.run("create protobuf")
        c.run("create cmake")
        c.run("create m4")
        c.run("create app --format=json", redirect_stdout="graph.json")
        # Now, let's filter the pkgs
        c.run(f"list --graph=graph.json --graph-context={context} --format=json",
              redirect_stdout="pkglist.json")
        content = json.loads(c.load("pkglist.json"))
        for ref in refs:
            assert content["Local Cache"][ref]["revisions"]
        for ref in not_refs:
            assert content["Local Cache"].get(ref) is None
        # Let's upload all the packages
        c.run(f"list --graph=graph.json --format=json", redirect_stdout="pkglist.json")
        c.run("upload --list=pkglist.json -r=default")
        c.run("remove * -c")
        c.run("create app --format=json", redirect_stdout="graph.json")
        # Now, let's filter the pkgs
        c.run(f"list --graph=graph.json --graph-context={context} -r=default --format=json",
              redirect_stdout="pkglist.json")
        content = json.loads(c.load("pkglist.json"))
        for ref in refs:
            assert content["default"][ref]["revisions"]
        for ref in not_refs:
            assert content["default"].get(ref) is None

    def test_graph_list(self):
        tc = TestClient(light=True)
        tc.save({"conanfile.py": GenConanfile("lib", "1.0")})

        tc.run("graph info . -f json --build=never --no-remote --filter=context",
               redirect_stdout="graph_context.json")

        tc.run("list --graph=graph_context.json --graph-context=build-only --format=json",
               assert_error=True)
        assert "Note that the graph file should not be filtered" in tc.out



import os
import platform

import pytest
import textwrap

from conan.internal.api.new.cmake_lib import source_cpp, source_h
from conan.internal.api.new.qbs_lib import qbs_lib_file
from conan.test.utils.tools import TestClient
from jinja2 import Template


def gen_file(template, **context):
    t = Template(template)
    return t.render(**context)


@pytest.mark.parametrize('shared', [False, True])
@pytest.mark.skipif(platform.system() != "Linux", reason="QBS only for Linux now")
@pytest.mark.tool("qbs")
def test_api_qbs_create_lib(shared):
    client = TestClient()
    client.run("new qbs_lib -d name=hello -d version=1.0")
    client.run("create . -o:h &:shared={shared}".format(shared=shared))
    assert "compiling hello.cpp" in client.out


@pytest.mark.skipif(platform.system() != "Linux", reason="QBS only for Linux now")
@pytest.mark.tool("qbs")
def test_qbs_all_products():
    client = TestClient()

    context = {
        "name": "hello",
        "version": "2.0",
        "package_name": "hello"
    }

    conanfile = textwrap.dedent('''
    import os

    from conan import ConanFile
    from conan.tools.qbs import Qbs

    class Recipe(ConanFile):
        name = "hello"
        version = "1.0"

        exports_sources = "*.cpp", "*.h", "*.qbs"
        settings = "os", "compiler", "arch", "build_type"

        def build(self):
            qbs = Qbs(self)
            qbs.resolve()
            qbs.build_all()
        ''')

    client.save({
        "conanfile.py": conanfile,
        "hello.cpp": gen_file(source_cpp, **context),
        "hello.h": gen_file(source_h, **context),
        "hello.qbs": gen_file(qbs_lib_file, **context),
    }, clean_first=True)

    client.run("create .")
    assert "--all-products" in client.out


@pytest.mark.skipif(platform.system() != "Linux", reason="QBS only for Linux now")
@pytest.mark.tool("qbs")
def test_qbs_specific_products():
    client = TestClient()

    context = {
        "name": "hello",
        "version": "2.0",
        "package_name": "hello"
    }

    conanfile = textwrap.dedent('''
    import os

    from conan import ConanFile
    from conan.tools.qbs import Qbs

    class Recipe(ConanFile):
        name = "hello"
        version = "1.0"

        exports_sources = "*.cpp", "*.h", "*.qbs"
        settings = "os", "compiler", "arch", "build_type"

        def build(self):
            qbs = Qbs(self)
            qbs.resolve()
            qbs.build(products=["hello", "hello"])
        ''')

    client.save({
        "conanfile.py": conanfile,
        "hello.cpp": gen_file(source_cpp, **context),
        "hello.h": gen_file(source_h, **context),
        "hello.qbs": gen_file(qbs_lib_file, **context),
    }, clean_first=True)

    client.run("create .")
    assert "--products hello,hello" in client.out


@pytest.mark.skipif(platform.system() != "Linux", reason="QBS only for Linux now")
@pytest.mark.tool("qbs")
def test_qbs_multiple_configurations():
    client = TestClient()

    context = {
        "name": "hello",
        "version": "2.0",
        "package_name": "hello"
    }

    conanfile = textwrap.dedent('''
    import os

    from conan import ConanFile
    from conan.tools.qbs import Qbs

    class Recipe(ConanFile):
        name = "hello"
        version = "1.0"

        exports_sources = "*.cpp", "*.h", "*.qbs"
        settings = "os", "compiler", "arch"

        def build(self):
            qbs = Qbs(self)
            qbs.add_configuration("release", {"qbs.debugInformation": False})
            qbs.add_configuration("debug", {"qbs.debugInformation": True})
            qbs.resolve()
            qbs.build()
        ''')

    client.save({
        "conanfile.py": conanfile,
        "hello.cpp": gen_file(source_cpp, **context),
        "hello.h": gen_file(source_h, **context),
        "hello.qbs": gen_file(qbs_lib_file, **context),
    }, clean_first=True)

    client.run("create .")
    build_folder = client.created_layout().build()
    assert os.path.exists(os.path.join(build_folder, "release"))
    assert os.path.exists(os.path.join(build_folder, "debug"))



import os
import shutil

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_db_error():
    # https://github.com/conan-io/conan/issues/14517
    c = TestClient(default_server_user=True)
    c.save({"liba/conanfile.py": GenConanfile("liba", "0.1")})
    c.run("create liba")
    c.run("install --requires=liba/0.1 --format=json", redirect_stdout="graph.json")
    c.run("list --graph=graph.json --format=json", redirect_stdout="installed.json")
    c.run("upload --list=installed.json -r=default --format=json -c", redirect_stdout="upload.json")

    c2 = TestClient(servers=c.servers, inputs=["admin", "password"])
    shutil.copy(os.path.join(c.current_folder, "upload.json"), c2.current_folder)
    c2.run("download --list=upload.json -r=default --format=json")
    # This used to crash
    assert "liba/0.1: Downloaded package revision" in c2.out



import re

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient, TestRequester


def test_offline():
    c = TestClient(servers={"default": None}, requester_class=None)
    c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
    c.run("create .")
    c.run("install --requires=pkg/0.1")
    # doesn't fail, install without issues
    assert "Install finished successfully" in c.out


def test_offline_uploaded():
    c = TestClient(default_server_user=True)
    c.save({"conanfile.py": GenConanfile("pkg")})
    c.run("create . --version=0.1")
    c.run("create . --version=0.2")
    c.run("upload * -r=default -c")
    c.run("remove * -c")
    c.run("install --requires=pkg/0.1")
    assert "Install finished successfully" in c.out
    c.servers = {"default": None}
    c.run("install --requires=pkg/0.1")
    assert "Install finished successfully" in c.out
    # Lets make sure the server is broken
    c.run("install --requires=pkg/0.2", assert_error=True)
    assert "ERROR: Package 'pkg/0.2' not resolved" in c.out


def test_offline_build_requires():
    """
    When there are tool-requires that are not installed locally, Conan will try to check
    its existence in the remote, because that package might be needed later. Even if some
    packages can be marked later as "skip" and not fail, that cannot be known a priori, so if the
    package is not in the cache, it will be checked in servers
    https://github.com/conan-io/conan/issues/15339

    Approaches to avoid the WARNING:
    - conan remote disable <remotes>
    - conan install ... -nr (--no-remotes)
    - prepopulate the cache with all tool-requires with `-c:a tools.graph:skip_binaries=False`
    """
    c = TestClient(default_server_user=True)
    c.save({"tool/conanfile.py": GenConanfile("tool", "0.1"),
            "lib/conanfile.py": GenConanfile("pkg", "0.1").with_tool_requires("tool/0.1")})
    c.run("create tool")
    c.run("create lib")
    c.run("upload * -r=default -c")
    c.run("remove * -c")
    c.run("install --requires=pkg/0.1")
    assert "Install finished successfully" in c.out

    class MyHttpRequester(TestRequester):

        def get(self, _, **kwargs):
            from requests.exceptions import ConnectionError
            raise ConnectionError("ALL BAD")

    c.requester_class = MyHttpRequester
    # this will fail
    c.run("install --requires=pkg/0.1", assert_error=True)
    assert "ERROR: Failed checking for binary 'tool/0.1:da39a3ee5e6b4b0d3255bfef95601890afd80709'" \
           in c.out
    # Explicitly telling that no-remotes, works
    c.run("install --requires=pkg/0.1 -nr")
    assert "tool/0.1: WARN" not in c.out
    assert "Install finished successfully" in c.out

    # graph info also breaks
    c.run("graph info --requires=pkg/0.1", assert_error=True)
    assert "ERROR: Failed checking for binary 'tool/0.1:da39a3ee5e6b4b0d3255bfef95601890afd80709'" \
           in c.out

    c.run("graph info --requires=pkg/0.1 -nr")
    assert "tool/0.1: WARN" not in c.out
    assert re.search(r"Skipped binaries(\s*)tool/0.1", c.out)



import textwrap

from unittest import mock

from conan import __version__
from conan.test.utils.tools import TestClient


def test_required_conan_version():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        required_conan_version = ">=100.0"

        class Lib(ConanFile):
            pass
        """)
    client.save({"conanfile.py": conanfile})
    client.run("export . --name=pkg --version=1.0", assert_error=True)
    assert f"Current Conan version ({__version__}) does not satisfy the defined one (>=100.0)" in client.out
    client.run("source . ", assert_error=True)
    assert f"Current Conan version ({__version__}) does not satisfy the defined one (>=100.0)" in client.out

    with mock.patch("conan.__version__", "101.0"):
        client.run("export . --name=pkg --version=1.0")

    with mock.patch("conan.__version__", "101.0-dev"):
        client.run("export . --name=pkg --version=1.0")

    client.run("install --requires=pkg/1.0@", assert_error=True)
    assert f"Current Conan version ({__version__}) does not satisfy the defined one (>=100.0)" in client.out


def test_required_conan_version_with_loading_issues():
    # https://github.com/conan-io/conan/issues/11239
    client = TestClient()
    conanfile = textwrap.dedent("""
                from conan import missing_import

                required_conan_version = ">=100.0"

                class Lib(ConanFile):
                    pass
                """)
    client.save({"conanfile.py": conanfile})
    client.run("export . --name=pkg --version=1.0", assert_error=True)
    assert f"Current Conan version ({__version__}) does not satisfy the defined one (>=100.0)" in client.out

    # Assigning required_conan_version without spaces
    conanfile = textwrap.dedent("""
                        from conan import missing_import

                        required_conan_version=">=100.0"

                        class Lib(ConanFile):
                            pass
                        """)
    client.save({"conanfile.py": conanfile})
    client.run("export . --name=pkg --version=1.0", assert_error=True)
    assert f"Current Conan version ({__version__}) does not satisfy the defined one (>=100.0)" in client.out

    # If the range is correct, everything works, of course
    conanfile = textwrap.dedent("""
                        from conan import ConanFile

                        required_conan_version = ">1.0.0"

                        class Lib(ConanFile):
                            pass
                        """)
    client.save({"conanfile.py": conanfile})
    client.run("export . --name=pkg --version=1.0")
    assert "pkg/1.0: Exported" in client.out


def test_comment_after_required_conan_version():
    """
    An error used to pop out if you tried to add a comment in the same line than
    required_conan_version, as it was trying to compare against >=10.0 # This should work
    instead of just >= 10.0
    """
    client = TestClient()
    conanfile = textwrap.dedent("""
                from conan import ConanFile
                from LIB_THAT_DOES_NOT_EXIST import MADE_UP_NAME
                required_conan_version = ">=10.0" # This should work
                class Lib(ConanFile):
                    pass
                """)
    client.save({"conanfile.py": conanfile})
    client.run("export . --name=pkg --version=1.0", assert_error=True)
    assert f"Current Conan version ({__version__}) does not satisfy the defined one (>=10.0)" in client.out


def test_commented_out_required_conan_version():
    """
    Used to not be able to comment out required_conan_version if we had to fall back
    to regex check because of an error importing the recipe
    """
    client = TestClient()
    conanfile = textwrap.dedent("""
                from conan import ConanFile
                from LIB_THAT_DOES_NOT_EXIST import MADE_UP_NAME
                required_conan_version = ">=1.0" # required_conan_version = ">=100.0"
                class Lib(ConanFile):
                    pass
                """)
    client.save({"conanfile.py": conanfile})
    client.run("export . --name=pkg --version=10.0", assert_error=True)
    assert f"Current Conan version ({__version__}) does not satisfy the defined one (>=1.0)" not in client.out

    client = TestClient()
    conanfile = textwrap.dedent("""
                from conan import ConanFile
                from LIB_THAT_DOES_NOT_EXIST import MADE_UP_NAME
                # required_conan_version = ">=10.0"
                class Lib(ConanFile):
                    pass
                """)
    client.save({"conanfile.py": conanfile})
    client.run("export . --name=pkg --version=1.0", assert_error=True)
    assert f"Current Conan version ({__version__}) does not satisfy the defined one (>=10.0)" not in client.out


def test_required_conan_version_invalid_syntax():
    """ required_conan_version used to warn of mismatching versions if spaces were present,
     but now we have a nicer error"""
    # https://github.com/conan-io/conan/issues/12692
    client = TestClient()
    conanfile = textwrap.dedent("""
                from conan import ConanFile
                required_conan_version = ">= 1.0"
                class Lib(ConanFile):
                    pass""")
    client.save({"conanfile.py": conanfile})
    client.run("export . --name=pkg --version=1.0", assert_error=True)
    assert f"Current Conan version ({__version__}) does not satisfy the defined one (>= 1.0)" not in client.out
    assert 'Error parsing version range ">="' in client.out



from conan.test.utils.tools import TestClient, GenConanfile


class TestTransitiveIds:

    def test_transitive_library(self):
        # https://github.com/conan-io/conan/issues/6450
        client = TestClient(light=True)
        client.save_home({"global.conf": "core.package_id:default_unknown_mode=full_version_mode"})
        client.save({"conanfile.py": GenConanfile()})
        client.run("create . --name=liba --version=1.0")
        client.run("create . --name=liba --version=1.1")
        client.save({"conanfile.py": GenConanfile().with_require("liba/1.0")})
        client.run("create . --name=libb --version=1.0")
        client.save({"conanfile.py": GenConanfile().with_require("libb/1.0")})
        client.run("create . --name=libc --version=1.0")
        client.save({"conanfile.py": GenConanfile().with_require("libc/1.0")
                                                   .with_require("liba/1.0")})
        client.run("create . --name=libd --version=1.0")
        # The consumer forces to depend on liba/2, instead of liba/1
        client.save({"conanfile.py": GenConanfile().with_require("libc/1.0")
                                                   .with_requirement("liba/1.1", force=True)})
        client.run("create . --name=libd --version=1.0", assert_error=True)
        # both B and C require a new binary
        client.assert_listed_binary(
            {"liba/1.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache"),
             "libb/1.0": ("efdeb15efa0e3e6ce0c9ef9bd82c56e4b2188c8a", "Missing"),
             "libc/1.0": ("fea281081e89fcc36242a68b0dbb16276a84f624", "Missing"),
             "libd/1.0": ("13a3ed814d78c8b91f303db7d4cdd7c4e614323c", "Build")
             })

    def test_transitive_major_mode(self):
        # https://github.com/conan-io/conan/issues/6450
        # Test LibE->LibD->LibC->LibB->LibA
        # LibC declares that it only depends on major version changes of its upstream
        # So LibC package ID doesn't change, even if LibA changes
        # But LibD package ID changes, even if its direct dependency LibC doesn't
        client = TestClient(light=True)
        client.save_home({"global.conf": "core.package_id:default_unknown_mode=full_version_mode"})
        # LibA
        client.save({"conanfile.py": GenConanfile()})
        client.run("create . --name=liba --version=1.0")
        client.run("create . --name=liba --version=1.1")
        # libB -> LibA
        client.save({"conanfile.py": GenConanfile().with_require("liba/1.0")})
        client.run("create . --name=libb --version=1.0")
        # libC -> libB
        major_mode = "self.info.requires.major_mode()"
        client.save({"conanfile.py": GenConanfile().with_require("libb/1.0")
                                                   .with_package_id(major_mode)})
        client.run("create . --name=libc --version=1.0")
        # Check the LibC ref with RREV keeps the same
        client.assert_listed_binary({"libc/1.0": ("2aa3bb6726dfc6c825a91f3b465b769e770ee902",
                                                  "Build")})
        # LibD -> LibC
        client.save({"conanfile.py": GenConanfile().with_require("libc/1.0")})
        client.run("create . --name=libd --version=1.0")
        # LibE -> LibD, LibA/2.0
        client.save({"conanfile.py": GenConanfile().with_require("libd/1.0")
                                                   .with_requirement("liba/1.1", force=True)})
        client.run("create . --name=libe --version=1.0", assert_error=True)
        # Check the LibC ref with RREV keeps the same, it is in cache, not missing
        # But LibD package ID changes and is missing, because it depends transitively on LibA
        client.assert_listed_binary(
            {"liba/1.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache"),
             "libb/1.0": ("efdeb15efa0e3e6ce0c9ef9bd82c56e4b2188c8a", "Missing"),
             "libc/1.0": ("2aa3bb6726dfc6c825a91f3b465b769e770ee902", "Cache"),
             "libd/1.0": ("13a3ed814d78c8b91f303db7d4cdd7c4e614323c", "Missing"),
             "libe/1.0": ("f93a37dd8bc53991d7485235c6b3e448942c52ff", "Build")
             })

    def test_transitive_unrelated(self):
        # https://github.com/conan-io/conan/issues/6450
        client = TestClient(light=True)
        # LibA
        client.save({"conanfile.py": GenConanfile()})
        client.run("create . --name=liba --version=1.0")
        client.run("create . --name=liba --version=2.0")
        # libB -> LibA
        client.save({"conanfile.py": GenConanfile().with_require("liba/1.0")})
        client.run("create . --name=libb --version=1.0")
        # libC -> libB
        unrelated = "self.info.requires['libb'].unrelated_mode()"
        client.save({"conanfile.py": GenConanfile().with_require("libb/1.0")
                    .with_package_id(unrelated)})
        client.run("create . --name=libc --version=1.0")
        # LibD -> LibC
        client.save({"conanfile.py": GenConanfile().with_require("libc/1.0")})
        client.run("create . --name=libd --version=1.0")
        # LibE -> LibD, LibA/2.0
        client.save({"conanfile.py": GenConanfile().with_require("libd/1.0")
                    .with_requirement("liba/2.0", force=True)})
        client.run("create . --name=libe --version=1.0", assert_error=True)
        client.assert_listed_binary({"liba/2.0": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache"),
                                     "libb/1.0": ("9c5273fc489264e39c2c16494d8a4178f239e6e2", "Missing"),
                                     "libc/1.0": ("9c5273fc489264e39c2c16494d8a4178f239e6e2", "Missing"),
                                     "libd/1.0": ("060865748907651202e4caa396a4c11a11f91e28", "Missing")})

    def test_transitive_second_level_header_only(self):
        # https://github.com/conan-io/conan/issues/6450
        client = TestClient(light=True)
        # LibA
        client.save({"conanfile.py": GenConanfile()})
        client.run("create . --name=liba --version=1.0")
        client.run("create . --name=liba --version=2.0")
        # libB -> LibA
        client.save({"conanfile.py": GenConanfile().with_require("liba/1.0")})
        client.run("create . --name=libb --version=1.0")
        # libC -> libB

        unrelated = "self.info.clear()"
        client.save({"conanfile.py": GenConanfile().with_require("libb/1.0")
                                                   .with_package_id(unrelated)})
        client.run("create . --name=libc --version=1.0")
        # LibD -> LibC
        client.save({"conanfile.py": GenConanfile().with_require("libc/1.0")})
        client.run("create . --name=libd --version=1.0")
        client.assert_listed_binary({"libc/1.0": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache")})

        # LibE -> LibD, LibA/2.0
        client.save({"conanfile.py": GenConanfile().with_require("libd/1.0")
                                                   .with_requirement("liba/2.0", force=True)})
        client.run("create . --name=libe --version=1.0", assert_error=True)  # LibD is NOT missing!

        client.assert_listed_binary(
            {"liba/2.0": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache"),
             "libb/1.0": ("9c5273fc489264e39c2c16494d8a4178f239e6e2", "Missing"),
             "libc/1.0": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache"),
             "libd/1.0": ("060865748907651202e4caa396a4c11a11f91e28", "Missing")})

    def test_transitive_header_only(self):
        # https://github.com/conan-io/conan/issues/6450
        client = TestClient(light=True)
        client.save_home({"global.conf": "core.package_id:default_unknown_mode=full_version_mode"})
        client.save({"conanfile.py": GenConanfile()})
        client.run("create . --name=liba --version=1.0")
        client.run("create . --name=liba --version=2.0")
        client.save({"conanfile.py": GenConanfile().with_require("liba/1.0")
                                                   .with_package_id("self.info.clear()")})
        client.run("create . --name=libb --version=1.0")
        client.save({"conanfile.py": GenConanfile().with_require("libb/1.0")})
        client.run("create . --name=libc --version=1.0")
        client.save({"conanfile.py": GenConanfile().with_require("libc/1.0")
                                                   .with_require("liba/1.0")})
        client.run("create . --name=libd --version=1.0")
        client.save({"conanfile.py": GenConanfile().with_require("libc/1.0")
                                                   .with_requirement("liba/2.0", force=True)})
        # USE THE NEW FIXED PACKAGE_ID
        client.run("create . --name=libd --version=1.0", assert_error=True)
        client.assert_listed_binary({"liba/2.0": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache"),
                                     "libb/1.0": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache"),
                                     "libc/1.0": ("2d7000cf800f37e01880897faf74a366a21bebdc", "Missing"),
                                     })



import json
import textwrap

from conan.test.utils.tools import TestClient


def test_options():
    """ lockfiles no longer contains option or any other configuration information. Instead
    the ``graph build-order`` applying a lockfile will return the necessary options to build
    it in order
    """
    client = TestClient(light=True)
    ffmpeg = textwrap.dedent("""
        from conan import ConanFile
        class FfmpegConan(ConanFile):
            options = {"variation": ["standard", "nano"]}
            default_options = {"variation": "standard"}

            def build(self):
                self.output.info("Variation %s!!" % self.options.variation)
        """)

    variant = textwrap.dedent("""
        from conan import ConanFile
        class Meta(ConanFile):
            requires = "ffmpeg/1.0"
            default_options = {"ffmpeg/1.0:variation": "nano"}
        """)

    client.save({"ffmepg/conanfile.py": ffmpeg,
                 "variant/conanfile.py": variant})
    client.run("export ffmepg --name=ffmpeg --version=1.0")
    client.run("export variant --name=nano --version=1.0")
    client.run("lock create --requires=nano/1.0@ --build=*")
    client.run("graph build-order --requires=nano/1.0@ "
               "--lockfile-out=conan.lock --build=missing "
               "--format=json", redirect_stdout="build_order.json")

    json_file = client.load("build_order.json")
    to_build = json.loads(json_file)
    ffmpeg = to_build[0][0]
    ref = ffmpeg["ref"]
    options = " ".join(f"-o {option}" for option in ffmpeg["packages"][0][0]["options"])

    cmd = "install --requires={} --build={} {}".format(ref, ref, options)
    client.run(cmd)
    assert "ffmpeg/1.0: Variation nano!!" in client.out



import copy
from collections import OrderedDict

import pytest
from unittest.mock import patch

from conan.api.model import RecipeReference
from conans.server.revision_list import RevisionList
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient, TestServer, NO_SETTINGS_PACKAGE_ID


class TestUpdateFlows:
    @pytest.fixture(autouse=True)
    def _setup(self):
        self.liba = RecipeReference.loads("liba/1.0.0")

        servers = OrderedDict()
        for index in range(3):
            servers[f"server{index}"] = TestServer([("*/*@*/*", "*")], [("*/*@*/*", "*")],
                                                   users={"user": "password"})

        self.client = TestClient(servers=servers, inputs=3*["user", "password"])
        self.client2 = TestClient(servers=servers, inputs=3*["user", "password"])
        self.the_time = 0.0
        self.server_times = {}

    def _upload_ref_to_all_servers(self, ref, client):
        # we are patching the time all these revisions uploaded to the servers
        # will be older than the ones we create in local
        for index in range(3):
            self.the_time = self.the_time + 10
            self._upload_ref_to_server(ref, f"server{index}", client)

    def _upload_ref_to_server(self, ref, remote, client):
        # we are patching the time all these revisions uploaded to the servers
        # will be older than the ones we create in local
        self.server_times[remote] = self.the_time
        with patch.object(RevisionList, '_now', return_value=self.the_time):
            client.run(f"upload {ref} -r {remote} -c")

    def test_revision_fixed_version(self):
        # NOTES:
        # - When a revision is installed from a remote it takes the date from the remote, not
        # updating the date to the current time
        # - If we want to install the revision and create it with a new date use --update-date
        # (name to be decided)
        # - Revisions are considered inmutable: if for example we do a conan install --update of a
        # revision that is already in the cache, but has a newer date in the remote, we will not install
        # anything, just updating the date in the cache to the one in the remote, so if you want to
        # get what the remote has you have to re-install you will have to remove the local
        # package and install from server
        # - In conan 2.X no remote means search in all remotes

        # create a revision 0 in client2, client2 will have an older revision than all the servers
        self.client2.save({"conanfile.py": GenConanfile("liba", "1.0.0").with_build_msg("REV0")})
        self.client2.run("create .")

        # other revision created in client
        self.client.save({"conanfile.py": GenConanfile("liba", "1.0.0").with_build_msg("REV")})
        self.client.run("create .")

        self._upload_ref_to_all_servers("liba/1.0.0", self.client)

        # upload other revision 1 we create in client
        self.client.save({"conanfile.py": GenConanfile("liba", "1.0.0").with_build_msg("REV1")})
        self.client.run("create .")

        self._upload_ref_to_all_servers("liba/1.0.0", self.client)

        # NOW WE HAVE:
        # | CLIENT      | CLIENT2    | SERVER0    | SERVER1   | SERVER2    |
        # |-------------|------------|------------|-----------|------------|
        # | REV1  (1020)| REV0 (1000)| REV1(40)   | REV1(50)  | REV1 (60)  |
        # | REV   (1010)|            | REV (10)   | REV (20)  | REV  (30)  |
        # |             |            |            |           |            |

        # 1. TESTING WITHOUT SPECIFIC REVISIONS AND WITH NO REMOTES: "conan install --requires=liba/1.0.0"

        # client2 already has a revision for this recipe, don't install anything
        self.client2.run("install --requires=liba/1.0.0@")
        self.client2.assert_listed_require({"liba/1.0.0": "Cache"})
        assert "liba/1.0.0: Already installed!" in self.client2.out

        self.client.run("remove * -c")

        # | CLIENT      | CLIENT2    | SERVER0    | SERVER1   | SERVER2    |
        # |-------------|------------|------------|-----------|------------|
        # |             | REV0 (1000)| REV1(40)   | REV1(50)  | REV1 (60)  |
        # |             |            | REV (10)   | REV (20)  | REV  (30)  |
        # |             |            |            |           |            |

        self.client.run("install --requires=liba/1.0.0@")

        # will not find revisions for the recipe -> search remotes by order and install the
        # first match that is rev1 from server0
        # --> result: install rev from server0
        self.client.assert_listed_require({"liba/1.0.0": "Downloaded (server0)"})
        assert f"liba/1.0.0: Retrieving package {NO_SETTINGS_PACKAGE_ID}" \
               " from remote 'server0'" in self.client.out

        latest_rrev = self.client.cache.get_latest_recipe_revision(RecipeReference.loads("liba/1.0.0@"))
        # check that we have stored REV1 in client with the same date from the server0
        assert latest_rrev.timestamp == self.server_times["server0"]
        assert latest_rrev.timestamp == self.server_times["server0"]

        # | CLIENT      | CLIENT2    | SERVER0    | SERVER1   | SERVER2    |
        # |-------------|------------|------------|-----------|------------|
        # |REV1 (40)    | REV0 (1000)| REV1(40)   | REV1(50)  | REV1 (60)  |
        # |             |            | REV (10)   | REV (20)  | REV  (30)  |
        # |             |            |            |           |            |

        self.client.run("install --requires=liba/1.0.0@ --update")
        # It will first check all the remotes and
        # will find the latest revision: REV1 from server2 we already have that
        # revision but the date is newer
        # --> result: do not download anything, but update REV1 date in cache
        self.client.assert_listed_require({"liba/1.0.0": "Cache (Updated date) (server2)"})
        assert "liba/1.0.0: Already installed!" in self.client.out

        # now create a newer REV2 in server2 and if we do --update it should update the date
        # to the date in server0 and associate that remote but not install anything

        # we create a newer revision in client2
        self.client2.run("remove * -c")
        self.client2.save({"conanfile.py": GenConanfile("liba", "1.0.0").with_build_msg("REV2")})
        self.client2.run("create .")

        self.the_time = 100.0

        self._upload_ref_to_server("liba/1.0.0", "server2", self.client2)

        # | CLIENT      | CLIENT2    | SERVER0    | SERVER1   | SERVER2    |
        # |-------------|------------|------------|-----------|------------|
        # |REV1 (60)    | REV2 (1000)| REV1(100)  | REV1(50)  | REV2 (100) |
        # |             |            | REV (10)   | REV (20)  | REV1 (60)  |
        # |             |            |            |           | REV  (30)  |

        self.client.run("install --requires=liba/1.0.0@ --update")
        # --> result: Update date and server because server0 has a newer date
        latest_rrev = self.client.cache.get_latest_recipe_revision(self.liba)
        self.client.assert_listed_require({"liba/1.0.0": "Updated (server2)"})
        assert "liba/1.0.0: Downloaded package" in self.client.out
        assert latest_rrev.timestamp == self.server_times["server2"]

        # we create a newer revision in client
        self.client.run("remove * -c")
        self.client.save({"conanfile.py": GenConanfile("liba", "1.0.0").with_build_msg("REV2")})
        self.client.run("create .")
        self.client.run(f"remove {latest_rrev.repr_notime()} -c -r server2")

        # | CLIENT      | CLIENT2    | SERVER0    | SERVER1   | SERVER2    |
        # |-------------|------------|------------|-----------|------------|
        # |REV2 (2002)  | REV2 (2000)| REV1(100)  | REV1(50)  | REV1 (60)  |
        # |             |            | REV (10)   | REV (20)  | REV (30)   |
        # |             |            |            |           |            |

        self.client.run("install --requires=liba/1.0.0@")
        # we already have a revision for liba/1.0.0 so don't install anything
        # --> result: don't install anything
        assert "liba/1.0.0: Already installed!" in self.client.out

        self.client.run("install --requires=liba/1.0.0@ --update")
        # we already have a newer revision in the client
        # we will check all the remotes, find the latest revision
        # this revision will be oldest than the one in the cache
        # --> result: don't install anything
        self.client.assert_listed_require({"liba/1.0.0": "Newer"})
        assert "liba/1.0.0: Already installed!" in self.client.out

        # create newer revisions in servers so that the ones from the clients are older
        self.client.save({"conanfile.py": GenConanfile("liba", "1.0.0").with_build_msg("REV3")})
        self.client.run("create .")
        rev_to_upload = self.client.cache.get_latest_recipe_revision(self.liba)
        # the future
        self.the_time = 3000000000.0
        self._upload_ref_to_all_servers(repr(rev_to_upload), self.client)

        # | CLIENT      | CLIENT2    | SERVER0    | SERVER1   | SERVER2    |
        # |-------------|------------|------------|-----------|------------|
        # | REV2 (2002) | REV2 (2000)| REV3(3010) | REV3(3020)| REV3 (3030)|
        # |             |            | REV1(100)  | REV1(50)  | REV1 (60)  |
        # |             |            | REV (10)   | REV (20)  | REV  (30)  |
        # |             |            |            |           |            |

        self.client2.run("install --requires=liba/1.0.0@ --update")
        # now check for newer references with --update for client2 that has an older revision
        # when we use --update: first check all remotes (no -r argument) get latest revision
        # check if it is in cache, if it is --> stop, if it is not --> check date and install
        # --> result: install rev from server2
        self.client2.assert_listed_require({"liba/1.0.0": "Updated (server2)"})
        assert f"liba/1.0.0: Downloaded recipe revision {rev_to_upload.revision}" in self.client2.out
        assert f"liba/1.0.0: Retrieving package {NO_SETTINGS_PACKAGE_ID}" \
               " from remote 'server2'" in self.client2.out

        check_ref = RecipeReference.loads(str(rev_to_upload))  # without revision
        rev_to_upload = self.client2.cache.get_latest_recipe_revision(check_ref)
        assert rev_to_upload.timestamp == self.server_times["server2"]

        # | CLIENT      | CLIENT2    | SERVER0    | SERVER1   | SERVER2    |
        # |-------------|------------|------------|-----------|------------|
        # | REV2 (2000) | REV3 (3030)| REV3(3010) | REV3(3020)| REV3 (3030)|
        # |             | REV0 (1000)| REV1(100)  | REV1(50)  | REV1 (60)  |
        # |             |            | REV (10)   | REV (20)  | REV  (30)  |
        # |             |            |            |           |            |

        # TESTING WITH SPECIFIC REVISIONS AND WITH NO REMOTES: "conan install --requires=liba/1.0.0#rrev"
        # - In conan 2.X no remote means search in all remotes

        # check one revision we already have will not be installed
        # we search for that revision in the cache, we found it
        # --> result: don't install that
        latest_rrev = self.client.cache.get_latest_recipe_revision(self.liba)
        self.client.run(f"install --requires={latest_rrev}@#{latest_rrev.revision}")
        self.client.assert_listed_require({"liba/1.0.0": "Cache"})
        assert "liba/1.0.0: Already installed!" in self.client.out

        self.client.run("remove * -c")

        self.client.run("remove '*' -c -r server0")
        self.client.run("remove '*' -c -r server1")
        self.client.run("remove '*' -c -r server2")

        # create new older revisions in servers
        self.client.save({"conanfile.py": GenConanfile("liba", "1.0.0").with_build_msg("REV4")})
        self.client.run("create .")
        server_rrev = self.client.cache.get_latest_recipe_revision(self.liba)
        self.the_time = 0.0

        self._upload_ref_to_all_servers("liba/1.0.0", self.client)

        self.client.run("remove * -c")

        # | CLIENT      | CLIENT2    | SERVER0    | SERVER1   | SERVER2    |
        # |-------------|------------|------------|-----------|------------|
        # |             | REV3 (3030)| REV4(10)   | REV4(20)  | REV4 (30)  |
        # |             | REV0 (1000)|            |           |            |
        # |             |            |            |           |            |

        self.client.save({"conanfile.py": GenConanfile("liba", "1.0.0").with_build_msg("REV5")})
        self.client.run("create .")

        # | CLIENT      | CLIENT2    | SERVER0    | SERVER1   | SERVER2    |
        # |-------------|------------|------------|-----------|------------|
        # | REV5 (2001) | REV3 (3030)| REV4(10)   | REV4(20)  | REV4 (30)  |
        # |             | REV0 (1000)|            |           |            |
        # |             |            |            |           |            |

        # install REV4
        self.client.run(f"install --requires={server_rrev}@#{server_rrev.revision}")
        # have a newer different revision in the cache, but ask for an specific revision that is in
        # the servers, will try to find that revision and install it from the first server found
        # will not check all the remotes for the latest because we consider revisions completely
        # immutable so all of them are the same
        # --> result: install new revision asked, but the latest revision remains the other one,
        # because the one installed took the date from the server and it's older
        assert "liba/1.0.0: Not found in local cache, looking in remotes..." in self.client.out
        assert "liba/1.0.0: Checking remote: server0" in self.client.out
        assert "liba/1.0.0: Checking remote: server1" not in self.client.out
        assert "liba/1.0.0: Checking remote: server2" not in self.client.out
        server_rrev_norev = copy.copy(server_rrev)
        server_rrev_norev.revision = None
        latest_cache_revision = self.client.cache.get_latest_recipe_revision(server_rrev_norev)
        assert latest_cache_revision != server_rrev

        # | CLIENT      | CLIENT2    | SERVER0    | SERVER1   | SERVER2    |
        # |-------------|------------|------------|-----------|------------|
        # | REV5 (2001) | REV3 (3030)| REV4(10)   | REV4(20)  | REV4 (30)  |
        # | REV4 (10)   | REV0 (1000)|            |           |            |
        # |             |            |            |           |            |

        self.client.run(f"install --requires={server_rrev}@#{server_rrev.revision} --update")
        # last step without --update it took the REV4 from server0 but now
        # we tell conan to search for newer recipes of an specific revision
        # it will go to server2 and update the local date with the one
        # from the remote
        # --> result: update REV4 date to 30 but it won't be latest

        latest_cache_revision = self.client.cache.get_latest_recipe_revision(server_rrev_norev)
        assert latest_cache_revision != server_rrev
        latest_cache_revision = self.client.cache.recipe_layout(server_rrev).reference

        assert self.the_time == latest_cache_revision.timestamp
        self.client.assert_listed_require({"liba/1.0.0": "Cache (Updated date) (server2)"})

        self.client.run("remove * -c")
        self.client.run("remove '*' -c -r server0")
        self.client.run("remove '*' -c -r server1")
        self.client.run("remove '*' -c -r server2")

        self.client.save({"conanfile.py": GenConanfile("liba", "1.0.0").with_build_msg("REV6")})
        self.client.run("create .")
        server_rrev = self.client.cache.get_latest_recipe_revision(self.liba)
        self.the_time = 3000000020.0

        self._upload_ref_to_all_servers(repr(server_rrev), self.client)

        latest_server_time = self.the_time

        # | CLIENT      | CLIENT2    | SERVER0    | SERVER1   | SERVER2    |
        # |-------------|------------|------------|-----------|------------|
        # | REV6(2002)  | REV3 (3030)| REV6(3030) |REV6(3040) | REV6(3050) |
        # |             | REV0 (1000)|            |           |            |
        # |             |            |            |           |            |

        self.client.run(f"install --requires={server_rrev}@#{server_rrev.revision} --update")

        # now we have the same revision with different dates in the servers and in the cache
        # in this case, if we specify --update we will check all the remotes, if that revision
        # has a newer date in the servers we will take that date from the server but we will not
        # install anything, we are considering revisions fully immutable in 2.0
        # --> results: update revision date in cache, do not install anything

        latest_rrev_cache = self.client.cache.get_latest_recipe_revision(self.liba)
        assert latest_server_time == latest_rrev_cache.timestamp
        self.client.assert_listed_require({"liba/1.0.0": "Cache (Updated date) (server2)"})

        # | CLIENT      | CLIENT2    | SERVER0    | SERVER1   | SERVER2    |
        # |-------------|------------|------------|-----------|------------|
        # | REV6(2002)  | REV3 (3050)| REV6(3030) |REV6(3040) | REV6(3050) |
        # |             | REV0 (1000)|            |           |            |
        # |             |            |            |           |            |

        self.client.run("remove * -c")

        # | CLIENT      | CLIENT2    | SERVER0    | SERVER1   | SERVER2    |
        # |-------------|------------|------------|-----------|------------|
        # |             | REV3 (3050)| REV6(3030) |REV6(3040) | REV6(3050) |
        # |             | REV0 (1000)|            |           |            |
        # |             |            |            |           |            |

        self.client.run(f"install --requires={server_rrev}@#{server_rrev.revision} --update")

        # now we have the same revision with different dates in the servers and in the cache
        # in this case, if we specify --update we will check all the remotes and will install
        # the revision from the server that has the latest date
        # --> results: install from server2

        latest_rrev_cache = self.client.cache.get_latest_recipe_revision(self.liba)
        assert latest_server_time == latest_rrev_cache.timestamp
        self.client.assert_listed_require({"liba/1.0.0": "Downloaded (server2)"})

        # | CLIENT      | CLIENT2    | SERVER0    | SERVER1   | SERVER2    |
        # |-------------|------------|------------|-----------|------------|
        # | REV6(3050)  | REV3 (3050)| REV6(3030) |REV6(3040) | REV6(3050) |
        # |             | REV0 (1000)|            |           |            |
        # |             |            |            |           |            |

    def test_version_ranges(self):
        # create a revision 0 in client2, client2 will have an older revision than all the servers
        for minor in range(3):
            self.client2.save({"conanfile.py": GenConanfile("liba", f"1.{minor}.0").with_build_msg("REV0")})
            self.client2.run("create .")
            self.the_time = 10.0 + minor*10.0
            self._upload_ref_to_server(f"liba/1.{minor}.0", f"server{minor}", self.client2)

        self.client.save({"conanfile.py": GenConanfile("liba", "1.0.0").with_build_msg("REV0")})
        self.client.run("create .")

        # NOW WE HAVE:
        # | CLIENT         | CLIENT2        | SERVER0        | SERVER1        | SERVER2        |
        # |----------------|----------------|----------------|----------------|----------------|
        # | 1.0 REV0 (1000)| 1.0 REV0 (1000)| 1.0 REV0 (10)  | 1.1 REV0 (10)  | 1.2 REV0 (10)  |
        # |                | 1.1 REV0 (1000)|                |                |                |
        # |                | 1.2 REV0 (1000)|                |                |                |

        self.client.run("install --requires=liba/[>0.9.0]@")
        assert "liba/[>0.9.0]: liba/1.0.0" in self.client.out
        assert "liba/1.0.0: Already installed!" in self.client.out

        self.client.run("remove * -c")

        # | CLIENT         | CLIENT2        | SERVER0        | SERVER1        | SERVER2        |
        # |----------------|----------------|----------------|----------------|----------------|
        # |                | 1.0 REV0 (1000)| 1.0 REV0 (10)  | 1.1 REV0 (20)  | 1.2 REV0 (30)  |
        # |                | 1.1 REV0 (1000)|                |                |                |
        # |                | 1.2 REV0 (1000)|                |                |                |

        self.client.run("install --requires=liba/[>0.9.0]@")

        # will not find versions for the recipe in cache -> search remotes by order and install the
        # first match that is 1.0 from server0
        # --> result: install 1.0 from server0
        assert "liba/[>0.9.0]: liba/1.0.0" in self.client.out
        self.client.assert_listed_require({"liba/1.0.0": "Downloaded (server0)"})

        latest_rrev = self.client.cache.get_latest_recipe_revision(RecipeReference.loads("liba/1.0.0@"))
        assert latest_rrev.timestamp == self.server_times["server0"]

        # | CLIENT         | CLIENT2        | SERVER0        | SERVER1        | SERVER2        |
        # |----------------|----------------|----------------|----------------|----------------|
        # | 1.0 REV0 (10)| 1.0 REV0 (1000)  | 1.0 REV0 (10)  | 1.1 REV0 (20)  | 1.2 REV0 (30)  |
        # |                | 1.1 REV0 (1000)|                |                |                |
        # |                | 1.2 REV0 (1000)|                |                |                |

        self.client.run("install --requires=liba/[>1.0.0]@")
        # first match that is 1.1 from server1
        # --> result: install 1.1 from server1
        assert "liba/[>1.0.0]: liba/1.1.0" in self.client.out
        self.client.assert_listed_require({"liba/1.1.0": "Downloaded (server1)"})

        # | CLIENT         | CLIENT2        | SERVER0        | SERVER1        | SERVER2        |
        # |----------------|----------------|----------------|----------------|----------------|
        # | 1.1 REV0 (10)  | 1.0 REV0 (1000)| 1.0 REV0 (10)  | 1.1 REV0 (20)  | 1.2 REV0 (30)  |
        # |                | 1.1 REV0 (1000)|                |                |                |
        # |                | 1.2 REV0 (1000)|                |                |                |

        self.client.run("install --requires=liba/[>1.0.0]@ --update")
        # check all servers
        # --> result: install 1.2 from server2
        assert "liba/[>1.0.0]: liba/1.2.0" in self.client.out
        self.client.assert_listed_require({"liba/1.2.0": "Downloaded (server2)"})

        # If we have multiple revisions with different names for the same version and we
        # do a --update we are going to first resolver the version range agains server0
        # then in the proxy we will install rev2 that is the latest
        # | CLIENT         | CLIENT2        | SERVER0        | SERVER1        | SERVER2        |
        # |----------------|----------------|----------------|----------------|----------------|
        # | 1.0 REV0 (10)  |                | 1.2 REV0 (10)  | 1.2 REV1 (20)  | 1.2 REV2 (30)  |
        # |                |                |                |                |                |
        # |                |                |                |                |                |

        self.client.run("remove * -c")
        self.client2.run("remove * -c")

        # now we are uploading different revisions with different dates, but the same version
        for minor in range(3):
            self.client2.save({"conanfile.py": GenConanfile("liba", f"1.2.0").with_build_msg(f"REV{minor}")})
            self.client2.run("create .")
            self.the_time = 10.0 + minor*10.0
            self._upload_ref_to_server(f"liba/1.2.0", f"server{minor}", self.client2)

        self.client.save({"conanfile.py": GenConanfile("liba", "1.0.0").with_build_msg("REV0")})
        self.client.run("create .")

        self.client.run("install --requires=liba/[>1.0.0]@ --update")
        assert "liba/[>1.0.0]: liba/1.2.0" in self.client.out
        self.client.assert_listed_require({"liba/1.2.0": "Downloaded (server2)"})
        assert f"liba/1.2.0: Retrieving package {NO_SETTINGS_PACKAGE_ID} " \
               "from remote 'server2' " in self.client.out


@pytest.mark.parametrize("update,result", [
                                           # Not a real pattern, works to support legacy syntax
                                           ["*", {"liba/1.1": "Downloaded (default)",
                                                  "libb/1.1": "Downloaded (default)"}],
                                           ["libc", {"liba/1.0": "Cache",
                                                     "libb/1.0": "Cache"}],
                                           ["liba", {"liba/1.1": "Downloaded (default)",
                                                       "libb/1.0": "Cache"}],
                                           ["libb", {"liba/1.0": "Cache",
                                                       "libb/1.1": "Downloaded (default)"}],
                                           ["", {"liba/1.0": "Cache",
                                                 "libb/1.0": "Cache"}],
                                           # Patterns not supported, only full name match
                                           ["lib*", {"liba/1.0": "Cache",
                                                     "libb/1.0": "Cache"}],
                                           ["liba/*", {"liba/1.0": "Cache",
                                                       "libb/1.0": "Cache"}],
                                           # None only passes legacy --update without args,
                                           # to ensure it works, it should be the same as passing *
                                           [None, {"liba/1.1": "Downloaded (default)",
                                                   "libb/1.1": "Downloaded (default)"}]
                                           ])
def test_muliref_update_pattern(update, result):
    tc = TestClient(light=True, default_server_user=True)
    tc.save({"liba/conanfile.py": GenConanfile("liba"),
             "libb/conanfile.py": GenConanfile("libb")})
    tc.run("create liba --version=1.0")
    tc.run("create libb --version=1.0")

    tc.run("create liba --version=1.1")
    tc.run("create libb --version=1.1")

    tc.run("upload * -c -r default")
    tc.run('remove "*/1.1" -c')

    update_flag = f"--update={update}" if update is not None else "--update"

    tc.run(f'install --requires="liba/[>=1.0]" --requires="libb/[>=1.0]" -r default {update_flag}')

    tc.assert_listed_require(result)
